#cs
AAH - Audacity Audio Helper
AutoIt v3.3.14.5 - http://www.autoitscript.com

Copyright (c) 2020, Bernard Condrau.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------------------------------
This script was inspired by EAH (c) Bernard Condrau, 2020
- see https://wiki.condrau.com/media:eah and https://wiki.condrau.com/media:aah
- Audacity Audio Helper serves the same purpose as EAH for track files encoded with Audacity

Audacity Audio Helper was compiled with AutoIt v3.3.14.5
- see available place holders at the end of this file
----------------------------------------------------------------------------------------------------
The default settings for metaflac and metamp3
track (%title% is added by Audacity)
metamp3 : --2 --artist "%artist%" --album "%albumtitle%" --year %year% --frame "TRCK:%tracknr1%/%numtracks%" --frame "TCON:%genre%" --frame "TPE2:%albumartist%" --frame "TPE3:%albuminterpret%" --frame "TPOS:%cdnumber%/%totalcds%" --frame "TCOM:%albumcomposer%" "%workspace%\*.mp3"
metaflac: --set-tag=ARTIST="%artist%" --set-tag=ALBUM="%albumtitle%" --set-tag=DATE=%year% --set-tag=TRACKNUMBER=%tracknr1% --set-tag=TOTALTRACKS=%numtracks% --set-tag=GENRE="%genre%" --set-tag=ALBUMARTIST="%albumartist%" --set-tag=PERFORMER="%albuminterpret%" --set-tag=DISCNUMBER=%cdnumber% --set-tag=TOTALDISCS=%totalcds% --set-tag=COMPOSER="%composer%" --show-tag=TITLE "%workspace%\*.flac"
common
metamp3 : --2 --replay-gain --frame "APIC{3}:%workspace%\%coverfile%" --comment "%comment%" --2 "%workspace%\*.mp3"
metaflac: --add-replay-gain --import-picture-from="%workspace%\%coverfile%" --set-tag=COMMENT="%comment%" --show-tag=TITLE "%workspace%\*.flac"
----------------------------------------------------------------------------------------------------
Note on debugging this script
- Set Globals $debug = 1 and $breakPointExe
- #include <Debug.au3>
- $breakPointExe points to a local DOS command batch file on your computer with following content
  @ECHO OFF
  ECHO Break Point: %1
  PAUSE
- Whenever you call function breakPoint(), the script pauses until hit a key in the DOS batch File
  This helps trouble shoot file modification, copy, and move commands
----------------------------------------------------------------------------------------------------
#ce

#pragma compile(UPX, True)
#pragma compile(AutoItExecuteAllowed, False)
#pragma compile(FileVersion, 1.0.20.1130)
#pragma compile(ProductName, Audacity Audio Helper)
#pragma compile(ProductVersion, 1.0)
#pragma compile(LegalCopyright, (c) 2020 Bernard Condrau)
#pragma compile(Comments, GNU General Public License version 3)
#pragma compile(OriginalFilename, AAH.exe)

#include <GUIConstantsEx.au3>
#include <MsgBoxConstants.au3>
#include <Array.au3>
#include <File.au3>


; set debug=0 for final build
Global $debug = 0
Global $breakPointExe = '"C:\PortableCustom\EAH\tools\breakpoint.bat"'
;#include <Debug.au3>

; app
; ----------------------------------------------------------------------------------------------------
Global $appName    = "Audacity Audio Helper"
Global $appVersion = "1.0.20.1130" ; same as #pragma above
Global $appDate    = "30 Nov 2020"
Global $appHelpURL = "https://wiki.condrau.com/media:aah"
Global $appAbout   = " AAH..."

; globals
; ----------------------------------------------------------------------------------------------------
Global $scriptName = StringTrimRight(@ScriptName, 4)
Global $iniFile = @ScriptDir & "\" & $scriptName & ".ini"
Global $langFile = @ScriptDir & "\" & $scriptName & ".en.ini"
Global $logFileExt = "log" ; log file extension
Global $menuCnt, $menuHandle, $menuItem  = ""
Global $windowHandle, $windowTitle = $appName
Global $myMusicRegistry = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\User Shell Folders", "My Music")
Global $albumArtExeRegistry = RegRead("HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\AlbumArt.exe", "")
Global $albumArtOptPreset = '"%albumartist%" "%albumtitle%" ""%workspace%\%coverfile%"" /group local /autoclose'
Global $imageMagickExeRegistry = RegRead("HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\ImageMagick.exe", "")
Global $imageMagickOptPreset = '-depth 8 "%workspace%\%coverfile%" "%workspace%\%coverfile%"'
Global $aahCommentDefault = "Created with Audacity/" & $scriptName & ", %today%"
Global $vaNameDefault = "Various Artists"
Global $language,$workSpace,$aahComment,$vaName,$registerHotKeys,$notepadExe,$notepadExeWindows = "notepad.exe"
Global $metaExe,$metaOpt,$metaOptTrack,$metaOptRegex,$encodeExt,$encodePath,$encodeEnabled,$addMetaExe,$addMetaOpt,$addMetaOptTrack,$addMetaOptRegex,$addEncodeExt,$addEncodePath,$addEncodeEnabled
Global $albumArtPath,$albumArtExe,$albumArtOpt,$albumArtFile,$imageMagickExe,$imageMagickOpt,$saveCoverImage,$resizeCoverImage
Global $trackArtistsFile = "track_artists.txt"
Global $metaArtist, $vaSet = 0

; debugging
; ----------------------------------------------------------------------------------------------------
If $debug Then
	; avoid compile error missing function _DebugSetup()
	$debugger = "_DebugSetup"
	Call($debugger)
	$debugger = "_DebugReport"
	Call($debugger, "AutoIt runs x64 = " & @AutoItX64)
Else
	$debugger = "noDebugReport"
EndIf

Func noDebugReport($sData, $bLastError = False, $bExit = False)
   ; this function replaces _DebugReport when debugging is off, and does nothing
EndFunc

Func breakPoint($position = @scriptName)
   If $debug Then RunWait($breakPointExe & " " & $position)
EndFunc

; helpers
; ----------------------------------------------------------------------------------------------------
; callback for F1 and Help button
Func WM_HELP()
	ShellExecute($appHelpURL)
EndFunc

Func registerHotKey($cmd, $text = 0)
   If $registerHotKeys Then
	  Local $hotKey[5][3] = [ _
		 ["tagTracks", "+{F5}", @TAB & "Shift-F5"], _
		 ["getAlbumArt", "+{F9}", @TAB & "Shift-F9"], _
		 ["tagAndFile", "^{F9}", @TAB & "Ctrl-F9"], _
		 ["help", "", @TAB & "F1"], _ ; F1 is the predefined help key
		 ["exit", "", @TAB & "Alt-F4"] _ ; Alt-F4 is the predefined exit key
	  ]
	  For $i = 0 To UBound($hotKey, $UBOUND_ROWS) - 1
		 If $hotKey[$i][0] == $cmd Then Return $hotKey[$i][1+$text]
	  Next
   EndIf
   Return ""
EndFunc

Func insertMenu()

   Local $itemHandle, $itemText, $newItemText, $accessKey, $accessKeyPos

   ; set menu text and command IDs
   Global $command[8][3] = [ _
	  [lang("AAH_MENU_TRACK"), 0x2000, "tagTracks"], _
	  [lang("AAH_MENU_ART"), 0x2100, "getAlbumArt"], _
	  [lang("AAH_MENU_TAG"), 0x2200, "tagAndFile"], _
	  [lang("AAH_MENU_MULTI"), 0x2300, "multiCdFileRename"], _
	  ["", -1], _
	  [lang("AAH_MENU_ABOUT") & $appAbout, 0x2400, "about"], _
	  [lang("AAH_MENU_HELP"), 0x2500, "help"], _
	  [lang("AAH_MENU_EXIT"), 0x2600, "exit"] _
   ]

   Call($debugger, "Inserting menu")

   $menuHandle = GUICtrlCreateMenu(lang("AAH_MENU_ENCODE"))

   ; create menu and hotkeys
   For $i = 0 To UBound($command, $UBOUND_ROWS) - 1
	  $command[$i][0] = $command[$i][0] & registerHotKey($command[$i][2], 1)
	  $command[$i][1] = GUICtrlCreateMenuItem($command[$i][0], $menuHandle)
	  HotKeySet(registerHotKey($command[$i][2]), $command[$i][2])
   Next

   Call($debugger, "Inserting menu done")
EndFunc

; create the form fields in window
Func createFormFields()

   Call($debugger, "Create form fields")

   Global $formFieldArtist		; %albumartist%
   Global $formFieldAlbum		; %albumtitle%
   Global $formFieldComposer	; %albumcomposer%
   Global $formFieldPerformer	; %albuminterpret%
   Global $formFieldCDnumber	; %cdnumber%
   Global $formFieldTotalCDs	; %totalcds%
   Global $formFieldGenre		; %genre%
   Global $formFieldYear		; %year%
   Global $formFieldComment		; %comment%
   Global $formFieldVA			; "Various Artists"

   ; info text
   GUICtrlCreateLabel(lang("AAH_INFOTEXT"), 10, 13)

   ; album title
   Local $formFieldAlbumLabel = GUICtrlCreateLabel(lang("AAH_ALBUM"), 10, 53, 100)
   $formFieldAlbum = GUICtrlCreateInput('', 100, 50, 165)

   ; album artist
   Local $formFieldArtistLabel = GUICtrlCreateLabel(lang("AAH_ARTIST"), 10, 78, 100)
   $formFieldArtist = GUICtrlCreateInput('', 100, 75, 165)

   ; performer
   Local $formFieldPerformerLabel = GUICtrlCreateLabel(lang("AAH_PERFORMER"), 10, 103, 100)
   $formFieldPerformer = GUICtrlCreateInput('', 100, 100, 165)

   ; Year
   Local $formFieldYearLabel = GUICtrlCreateLabel(lang("AAH_YEAR"), 275, 53, 100)
   $formFieldYear = GUICtrlCreateInput('', 345, 50, 120)

   ; Genre
   Local $formFieldGenreLabel = GUICtrlCreateLabel(lang("AAH_GENRE"), 275, 78, 100)
   $formFieldGenre = GUICtrlCreateInput('', 345, 75, 120)

   ; CD Number
   Local $formFieldCDnumberLabel = GUICtrlCreateLabel(lang("AAH_CDNUMBER"), 275, 103, 100)
   $formFieldCDnumber = GUICtrlCreateInput('1', 345, 100, 40)

   ; composer
   Local $formFieldComposerLabel = GUICtrlCreateLabel(lang("AAH_COMPOSER"), 475, 53, 100)
   $formFieldComposer = GUICtrlCreateInput('', 565, 50, 165)

   ; Comment
   Local $formFieldCommentLabel = GUICtrlCreateLabel(lang("AAH_COMMENT"), 475, 78, 100)
   $formFieldComment = GUICtrlCreateInput('', 565, 75, 165)

   ; total CDs
   Local $formFieldTotalCDsLabel = GUICtrlCreateLabel(lang("AAH_TOTALCDS"), 475, 103, 100)
   $formFieldTotalCDs = GUICtrlCreateInput('1', 565, 100, 40)

   ; various artists checkbox
   $formFieldVA = GUICtrlCreateCheckbox(lang("AAH_MENU_SET_VA"), 100, 128)
EndFunc

Func getFormFields()

   Call($debugger, "Get form field values")

   Global $albumCDnumber, $albumTotalCDs ; we need these to handle CD numbering
   Global $albumArtist, $albumTitle, $albumComposer, $albumYear
   Local  $albumPerformer, $albumComment, $albumGenre, $today

   ; get form fields
   $albumTitle = GUICtrlRead($formFieldAlbum) ; mandatory
   $albumArtist = GUICtrlRead($formFieldArtist) ; mandatory
   $albumComposer = GUICtrlRead($formFieldComposer) ; cannot be empty
   $albumPerformer = GUICtrlRead($formFieldPerformer) ; cannot be empty
   $albumComment = GUICtrlRead($formFieldComment)
   $albumYear = GUICtrlRead($formFieldYear)
   $albumGenre = GUICtrlRead($formFieldGenre)
   $albumCDnumber = GUICtrlRead($formFieldCDnumber)
   $albumTotalCDs = GUICtrlRead($formFieldTotalCDs)
   $today = @YEAR & "-" & @MON & "-" & @MDAY ; must be file system conform

   ; we need to combine comments from the "Comment" form field and ini file
   If $albumComment == "" Then
	  $albumComment = $aahComment
   EndIf

   Global $albumInfo[13][3] = [ _
	  ["%coverfile%", $albumArtFile, replaceCharacters($albumArtFile)], _ ; must be first place holder as this might contain other place holders
	  ["%albumtitle%", $albumTitle, replaceCharacters($albumTitle)], _
	  ["%albumartist%", $albumArtist, replaceCharacters($albumArtist)], _
	  ["%albumcomposer%", $albumComposer, replaceCharacters($albumComposer)], _
	  ["%albuminterpret%", $albumPerformer, replaceCharacters($albumPerformer)], _
	  ["%comment%", $albumComment, $albumComment], _
	  ["%genre%", $albumGenre, $albumGenre], _
	  ["%cdnumber%", $albumCDnumber, $albumCDnumber], _
	  ["%totalcds%", $albumTotalCDs, $albumTotalCDs], _
	  ["%workspace%", $workSpace, $workSpace], _
	  ["%scriptdir%", @ScriptDir, @ScriptDir], _
	  ["%username%", @UserName, @UserName], _
	  ["%today%", $today, $today] _
   ]
   If $debug Then _ArrayDisplay($albumInfo)
EndFunc

Func iniFileRead($filename, $section, $key, $default, $exe = 0)

   Local $value

   $value = IniRead($filename, $section, $key, $default)
   If $value == "" Then $value = $default
   If $exe Then quotePath($value)
   Return $value

EndFunc

Func initialize()

   Global $dll = DllOpen("user32.dll")
   Local $myLanguage

   Call($debugger, "initialize")

   ; check whether my language translation file exists
   $myLanguage = IniFileRead($iniFile, "Settings", "Language", "en")
   If FileExists(@ScriptDir & "\" & $scriptName & "." & $myLanguage & ".ini") Then
	  $language = $myLanguage
	  $langFile = @ScriptDir & "\" & $scriptName & "." & $language & ".ini"
   EndIf

   ; read ini file - Settings
   $notepadExe = IniFileRead($iniFile, "Settings", "NotepadExe", $notepadExeWindows, 1)
   $workSpace = IniFileRead($iniFile, "Settings", "Workspace", $myMusicRegistry)
   $aahComment = IniFileRead($iniFile, "Settings", "Comment", $aahCommentDefault)
   $vaName = IniFileRead($iniFile, "Settings", "VA", $vaNameDefault)
   $registerHotKeys = Int(IniFileRead($iniFile, "Settings", "RegisterHotKeys", 0))

   ; read ini file - Encoders
   $metaExe = IniFileRead($iniFile, "Encoders", "MetaExe", "", 1)
   $metaOpt = IniFileRead($iniFile, "Encoders", "MetaOpt", "")
   $metaOptTrack = IniFileRead($iniFile, "Encoders", "MetaOptTrack", "")
   $metaOptRegex = IniFileRead($iniFile, "Encoders", "MetaOptRegex", "")
   $encodeExt = IniFileRead($iniFile, "Encoders", "EncodeExt", "")
   $encodePath = IniFileRead($iniFile, "Encoders", "EncodePath", "")
   $addMetaExe = IniFileRead($iniFile, "Encoders", "AddMetaExe", "", 1)
   $addMetaOpt = IniFileRead($iniFile, "Encoders", "AddMetaOpt", "")
   $addMetaOptTrack = IniFileRead($iniFile, "Encoders", "AddMetaOptTrack", "")
   $addMetaOptRegex = IniFileRead($iniFile, "Encoders", "AddMetaOptRegex", "")
   $addEncodeExt = IniFileRead($iniFile, "Encoders", "AddEncodeExt", "")
   $addEncodePath = IniFileRead($iniFile, "Encoders", "AddEncodePath", "")
   $encodeEnabled = $metaExe <> "" And $metaOpt <> "" And $encodeExt <> "" And $encodePath <> ""
   $addEncodeEnabled = $addMetaExe <> "" And $addMetaOpt <> "" And $addEncodeExt <> "" And $addEncodePath <> ""

   ; read ini file - AlbumArt
   $albumArtFile = IniFileRead($iniFile, "AlbumArt", "AlbumArtFile", "folder.jpg")
   $albumArtExe = IniFileRead($iniFile, "AlbumArt", "AlbumArtExe", $albumArtExeRegistry, 1)
   $albumArtOpt = IniFileRead($iniFile, "AlbumArt", "AlbumArtOpt", $albumArtOptPreset)
   $albumArtPath = IniFileRead($iniFile, "AlbumArt", "AlbumArtPath", "")
   $imageMagickExe = IniFileRead($iniFile, "AlbumArt", "ImageMagickExe", $imageMagickExeRegistry, 1)
   $imageMagickOpt = IniFileRead($iniFile, "AlbumArt", "ImageMagickOpt", $imageMagickOptPreset)
   $saveCoverImage = $albumArtPath <> ""
   $resizeCoverImage = $imageMagickExe <> "" And $imageMagickOpt <> ""

	; read character replacements table from registry and/or ini file
	getCharacterReplacements()

   ; remove trailing \
   trimSlash($workSpace)
   trimSlash($encodePath)
   trimSlash($addEncodePath)
   trimSlash($albumArtPath)

   Call($debugger, 'Reading ' & $iniFile & '...' & @CRLF & $language & ' (' & lang("AAH_LANGUAGE") & ') - ' & $langFile & @CRLF & $notepadExe & @CRLF & $albumArtExe & @CRLF & $albumArtFile & @CRLF & $workSpace & @CRLF & $encodePath & @CRLF & $addEncodePath)
   Call($debugger, $albumArtPath & @CRLF & $aahComment & @CRLF & $vaName & @CRLF & "Register hotkeys=" & $registerHotKeys)


EndFunc

Func lang($key)
   $str = IniRead($langFile, $language, $key, $key)
   Return StringReplace(StringReplace($str, "%TAB%", @TAB), "%CRLF%", @CRLF)
EndFunc

Func terminate()
   Call($debugger, 'terminate')
   DllClose($dll)
EndFunc

; file name manipulation
; ----------------------------------------------------------------------------------------------------
Func checkExists($path)
	Local $langMsg
	$path = StringRegExpReplace($path, '^"|"$', "")
	If $path <> "" And Not FileExists($path) Then
		$langMsg = "AAH_PATH_NOT_FOUND"
		If StringRegExp($path, '.*\.exe$') Then $langMsg = "AAH_FILE_NOT_FOUND"
		MsgBox($MB_OK + $MB_ICONWARNING, $path, lang($langMsg))
		Return False
	EndIf
	Return True
EndFunc

; get base name without file extension
Func baseName($name)
	Local $parts = StringSplit($name, ".")
	Local $ext = _ArrayPop($parts)
	Local $skip = $ext <> $encodeExt And $ext <> $addEncodeExt
	If $skip Then
		Return $name
	Else
		Return _ArrayToString($parts, ".", 1)
	EndIf
EndFunc

; handle audacity file names
; remove "-" between track number and filename
Global Const $regexAudacityPattern = '^([\d]{2,3})-([^\s])'			; start with "01-" or "001-" not followed by space
Global Const $regexMultiCdsPattern = '^([\d]{2,3}-[\d]{2,3}\s)'		; "01-01 " or "001-01 " are not audacity patterns, these are multi CDs
Global Const $regexAudacityReplace = '$1 $2'						; replacement pattern (01-filename => 01 filename, 001-filename => 001 filename)

; check whether file name was created by Audacity
Func checkAudacity($str)
	Return StringRegExp($str, $regexAudacityPattern) And Not StringRegExp($str, $regexMultiCdsPattern)
EndFunc

; rename Audacity file name to file name which can be processed
Func renameAudacity($str)
	Return StringStripWS(StringRegExpReplace($str, $regexAudacityPattern, $regexAudacityReplace), $STR_STRIPLEADING + $STR_STRIPTRAILING)
EndFunc

; check and rename all Audacity file names
Func renameAudacityFileNames($path, $ext)

   Local $id, $fileName, $fileHandle

   $fileHandle = FileFindFirstFile($path & "\*." & $ext)
   If $fileHandle <> -1 Then
	  While 1
		 $fileName = FileFindNextFile($fileHandle)
		 If @error Then ExitLoop

		 Call($debugger, "Renaming " & $fileName & "...")

		 If checkAudacity($fileName) Then
			If Not FileMove($path & "\" & $fileName, $path & "\" & renameAudacity($fileName)) Then
			   $id = MsgBox($MB_OKCANCEL + $MB_ICONWARNING, $path & "\" & $fileName, lang("AAH_AUDACITY_RENAME_ERR"))
			   If $id <> $IDOK Then Return False
			Else
			   Call($debugger, "File " & $path & "\" & $fileName & " renamed to " & renameAudacity($fileName))
			EndIf
		 EndIf
	  WEnd
   EndIf
   FileClose($fileHandle)
   Return True
EndFunc

; verify and rename valid file names
; check whether audacity file names have been renamed
Global Const $regexProcessPattern = '^([\d]{2,3})\s([^\s])'		; start with "01 " or "001 " not followed by another space
Global Const $regexProcessReplace = '$2'						; replacement pattern (01 filename => filename, 001 filename => filename)

; check whether file name can be processed
Func checkTrackFile($str)
	Return StringRegExp($str, $regexProcessPattern)
EndFunc

; remove track number from file name without extension
Func trimTrackNum($str)
	Return StringStripWS(StringRegExpReplace($str, $regexProcessPattern, $regexProcessReplace), $STR_STRIPLEADING + $STR_STRIPTRAILING)
EndFunc

; check whether audacity file names have been renamed
Func checkTracksReady($path, $ext)

   Local $fileName, $fileHandle
   Local $clean = -1 ; -1 = no files found, 0 = one or more files not renamed, 1 = all files ok

   $fileHandle = FileFindFirstFile($path & "\*." & $ext)
   If $fileHandle <> -1 Then
	  $clean = 0
	  While 1
		 $fileName = FileFindNextFile($fileHandle)
		 If @error Then
			$clean = 1
			ExitLoop
		 EndIf

		 ; check whether pattern exists, which means there is a file which has not been renamed
		 If checkAudacity($fileName) Then ExitLoop
	  WEnd
   EndIf
   FileClose($fileHandle)

   Return $clean
EndFunc

; verify info file
Global Const $regexInfoPattern = '^([\d]{1,2})-([^\s])'			; start with "1-" or "01-" not followed by space

; check whether info file can be processed
Func checkInfoFile($str)
	Return Not StringRegExp($str, $regexInfoPattern)
EndFunc

; remove more than 1 dot before the extension (metamp3.exe cannot handle it)
Func trimDots($path, $ext)

	Local $filteredName, $fileName, $fileHandle

	If $ext<>"" Then
		$fileHandle = FileFindFirstFile($path & "\*." & $ext)
		While 1
			$fileName = FileFindNextFile($fileHandle)
			If @error Then ExitLoop

			$filteredName = StringStripWS(StringRegExpReplace($fileName, "[.]{2,}" & $ext, "." & $ext), $STR_STRIPLEADING + $STR_STRIPTRAILING)
			If $filteredName <> $fileName Then FileMove($path & "\" & $fileName, $path & "\" & $filteredName)
		WEnd
	EndIf
EndFunc

; read character replacements table from EAC registry or ini file
Func getCharacterReplacements()

	Global $characterTable[0][2]
	Local $slash = False
	Local $empty = 16 ; add "/" to the end of the table if no table entry is empty

	For $i = 0 To 15
		$rule = IniFileRead($iniFile, "CharacterReplacements", "ReplacementRule" & $i, RegRead("HKEY_CURRENT_USER\Software\AWSoftware\EACU\Extraction Options", "ReplacementRule" & $i), 0)
		If $rule <> "" Then
			If StringLeft($rule, 1) == "/" Then $slash = True
			_ArrayAdd($characterTable, StringLeft($rule, 1) & "/?/" & StringTrimLeft($rule, 1), 0, "/?/")
		Else
			If $empty > 15 Then $empty = $i
			_ArrayAdd($characterTable, "")
		EndIf
	Next
	If Not $slash Then
		$characterTable[$empty][0] = "/"
		$characterTable[$empty][1] = ","
	EndIf
	If $debug Then _ArrayDisplay($characterTable)
EndFunc

; remove illegal file name characters
Func replaceCharacters($str)

	; replace characters
	For $i = 0 To UBound($characterTable) - 1
		If $characterTable[$i][0] <> "" Then $str = StringReplace($str, $characterTable[$i][0], $characterTable[$i][1], 0, $STR_CASESENSE)
	Next

	; remove preceding spaces
	While StringLeft($str, 1) == ' '
		$str = StringTrimLeft($str, 1)
	Wend

	; remove trailing spaces or dots
	While StringRight($str, 1) == ' ' Or StringRight($str, 1) == '.'
		$str = StringTrimRight($str, 1)
	Wend

	Return $str
EndFunc

; trim trailing slash
Func trimSlash(ByRef $path)
   If StringRight($path, 1) == '\' Then $path = StringTrimRight($path, 1)
EndFunc

; AutoIt's Run command requires quoted path/exe names
Func quotePath(ByRef $path)
   If StringLeft($path, 1) == '"' And StringRight($path, 1) == '"' Then Return
   If $path == "" Then Return
   $path = '"' & $path & '"'
EndFunc

; replace place holders
;	$str: string with place holders
;	$table: replacement table with patterns and replacement Strings. Col0=pattern, Col1=replacement string, Col2=filtered replacement string.
;	$regex: ""=no further processing, "filter"=replace illegal (file system) characters, "<regex string>"=regex pattern to replace with an empty string. Separate several regex patterns with " ".
;	Return: filtered and replaced string
; Note: if you need to filter and remove regex patterns, call the function twice
Func replacePlaceHolders($str, Const ByRef $table, $regex = "filter")

   Local $col = 1 ; column not filtered
   If $regex=="filter" Then $col = 2 ; column filtered

   ; replace place holders
   For $i = 0 to UBound($table, $UBOUND_ROWS) - 1
	  $str = StringReplace($str, $table[$i][0], $table[$i][$col])
   Next

   ; remove regex patterns
   If $regex<>"raw" And $regex<>"filter" Then
	 $pattern = StringSplit($regex, " ")
	 For $i = 1 to $pattern[0]
		$str = StringRegExpReplace($str, $pattern[$i], "")
	 Next
   EndIf

   Return $str
EndFunc

; progress window
; ----------------------------------------------------------------------------------------------------
Func progress($percent, $track = 0, $trackCount = 0)

	Static Local $previous = -1

	If $percent == 0 Then

		; calculate position of progress window
		Local $winPos = WinGetPos($windowHandle)
		Local $desktop = WinGetPos("Program Manager")
		Local $xPos = $winPos[0] + $winPos[2]/2 - 150
		Local $yPos = $winPos[1] - 120
		If $yPos < 0 Then $yPos = 0
		If $winPos[1] < 120 And $winPos[1] + $winPos[3] + 120 < $desktop[3] - 160 Then $yPos = $winPos[1] + $winPos[3] + 20

		; initialize progress window
		ProgressOn("", lang("AAH_PROGRESS"), lang("AAH_PROGRESS_PERCENT") & ": 0%" & @CRLF & lang("AAH_PROGRESS_TRACKS") & ": 0/0", $xPos, $yPos, BitOR($DLG_NOTITLE, $DLG_NOTONTOP))
	Else

		; update progress window
		$percent = Round($percent, 1)
		If $percent <> $previous Then
			ProgressSet($percent, lang("AAH_PROGRESS_PERCENT") & ": " & $percent & "%" & @CRLF & lang("AAH_PROGRESS_TRACKS") & ": " & $track & "/" & $trackCount)
			$previous = $percent
		EndIf

		; reset $previous
		If $previous == 100 Then $previous = -1
	EndIf
EndFunc

; actions
; ----------------------------------------------------------------------------------------------------
; about msg box
Func about()
	MsgBox($MB_ICONINFORMATION, lang("AAH_MENU_ABOUT") & " " & $appName, $appName & " v" & $appVersion & " (" & $appDate & ")" & @CRLF & $appHelpURL, 10)
EndFunc

; invoke help
Func help()
	WM_HELP()
EndFunc

; tag track specific meta tags
Func tagTracks()

   getFormFields()

   If Not $albumArtist Or Not $albumTitle Then
	  MsgBox($MB_OK + $MB_ICONERROR, lang("AAH_MANDATORY"), lang("AAH_MANDATORY_DESC"))
	  Return
   EndIf

   Local $phWorkSpace = replacePlaceHolders($workSpace, $albumInfo, "filter")
   Local $phMetaExe = replacePlaceHolders($metaExe, $albumInfo, "filter")
   Local $metaOptTrackTemplate = replacePlaceHolders($metaOptTrack, $albumInfo, "raw")
   Local $phAddMetaExe = replacePlaceHolders($addMetaExe, $albumInfo, "filter")
   Local $addMetaOptTrackTemplate = replacePlaceHolders($addMetaOptTrack, $albumInfo, "raw")
   Local $fileName, $fileHandle
   Local $id, $lines[0][1]
   Local $trackTitle, $trackNumber, $numTracks = 0
   Local $artist, $track, $year, $trackYear[1000], $trackArtist[1000], $trackComposer[1000]
   Local $trackInfo[6][2]

   ; consistency checks
   If Not checkExists($phWorkSpace) Then Return
   If Not checkExists($phMetaExe) Then Return
   If Not checkExists($phAddMetaExe) Then Return

	progress(0)

	; get track artists & composer from file
	If $encodeEnabled Or $addEncodeEnabled Then
	  $trackYear[0] = $albumYear ; fall back to $albumYear if line does not start with number or entry for track does not exist
	  $trackArtist[0] = $albumArtist ; fall back to $albumArtist if line does not start with number or entry for track does not exist
	  $trackComposer[0] = $albumComposer ; fall back to $albumArtist if line does not start with number or entry for track does not exist
	  If $vaSet Then
		 _FileReadToArray($phWorkSpace & "\" & $trackArtistsFile, $lines)
		 For $i = 1 to UBound($lines) -1
			$num = Number($lines[$i])
			If $num > 999 Then
			   $id = MsgBox($MB_YESNO + $MB_ICONQUESTION + $MB_DEFBUTTON2, lang("AAH_TRACKNUM_EXCEED"), lang("AAH_TRACKNUM_EXCEED_DEFINE_DESC"))
			   If $id <> $IDYES Then Return
			   $num = 999
			   ExitLoop
			EndIf
			If $num Then
			   $track = StringSplit($lines[$i], "**",  $STR_ENTIRESPLIT)
			   $year = $track[$track[0]]
			   If $track[0] > 1 And StringLen($year) == 4 And StringIsInt($year) Then
					$trackYear[$num] = $year
					$lines[$i] = StringTrimRight($lines[$i], 6)
			   EndIf
			   $track = StringSplit($lines[$i], "//",  $STR_ENTIRESPLIT)
			   If $track[0] == 1 Then $trackArtist[$num] = trimTrackNum($track[1])
			   If $track[0] == 2 Then
				  $trackArtist[$num] = trimTrackNum($track[1])
				  $trackComposer[$num] = $track[2]
			   EndIf
			EndIf
		 Next
	  EndIf
	EndIf

   ; tag files
   If $encodeEnabled Then
	  If Not checkTracksReady($phWorkSpace, $encodeExt) Then
		 If Not renameAudacityFileNames($phWorkSpace, $encodeExt) Then Return
	  EndIf
	  trimDots($phWorkSpace, $encodeExt)
	  $fileHandle = FileFindFirstFile($phWorkSpace & "\*." & $encodeExt)
	  If $fileHandle <> -1 Then

		 ; count Files
		 While 1
			$fileName = FileFindNextFile($fileHandle)
			If @error Then ExitLoop

			If checkTrackFile($fileName) Then $numTracks += 1
		 WEnd
		 FileClose($fileHandle)
		 Call($debugger, $numTracks & " files found")

		 If $numTracks > 999 Then
			$id = MsgBox($MB_YESNO + $MB_ICONQUESTION + $MB_DEFBUTTON2, lang("AAH_TRACKNUM_EXCEED"), lang("AAH_TRACKNUM_EXCEED_DESC"))
			If $id <> $IDYES Then Return
			$numTracks = 999
		 EndIf

		 $fileHandle = FileFindFirstFile($phWorkSpace & "\*." & $encodeExt)
		 While 1
			$fileName = FileFindNextFile($fileHandle)
			If @error Then ExitLoop

			If checkTrackFile($fileName) Then

				; get track number, track artist, track composer, track year
				$trackNumber = Number($fileName)
				$trackTitle = trimTrackNum(baseName($fileName))
				If $trackArtist[$trackNumber] Then
					$artist = $trackArtist[$trackNumber]
				Else
					$artist = $trackArtist[0]
				EndIf
				If $trackComposer[$trackNumber] Then
					$composer = $trackComposer[$trackNumber]
				Else
					$composer = $trackComposer[0]
				EndIf
				If $trackYear[$trackNumber] Then
					$year = $trackYear[$trackNumber]
				Else
					$year = $trackYear[0]
				EndIf

				; replace track specific place holders
				Local $trackInfo[9][2] = [ _
					["%tracknr%", StringFormat("%.2d", $trackNumber)], _
					["%tracknr1%", String($trackNumber)], _
					["%tracknr2%", StringFormat("%.2d", $trackNumber)], _
					["%tracknr3%", StringFormat("%.3d", $trackNumber)], _
					["%numtracks%", String($numTracks)], _
					["%year%", $year], _
					["%title%", $trackTitle], _
					["%composer%", $composer], _
					["%artist%", $artist] _
				]
				If $debug Then  _ArrayDisplay($trackInfo)

				; filter options, then call meta exe for each file (equivalent to EAC encoder call)
				$phMetaOptTrack = ReplacePlaceHolders($metaOptTrackTemplate, $trackInfo, $metaOptRegex)
				Call($debugger, 'Meta exe: ' & $phMetaExe & " " & $phMetaOptTrack & ' "' &  $phWorkSpace & "\" & $fileName & '"' & " (" & $metaOptTrack & ")")
				RunWait($phMetaExe & " " & $phMetaOptTrack & ' "' &  $phWorkSpace & "\" & $fileName & '"', "", @SW_HIDE)
				progress($trackNumber/$numTracks*100, $trackNumber, $numTracks)
			EndIf
		 WEnd
	  EndIf
	  FileClose($fileHandle)
	EndIf

   If $addEncodeEnabled Then
	  If Not checkTracksReady($phWorkSpace, $addEncodeExt) Then
		 If Not renameAudacityFileNames($phWorkSpace, $addEncodeExt) Then Return
	  EndIf
	  trimDots($phWorkSpace, $addEncodeExt)
	  $fileHandle = FileFindFirstFile($phWorkSpace & "\*." & $addEncodeExt)
	  If $fileHandle <> -1 Then

		 $numTracks = 0

		 ; count Files
		 While 1
			$fileName = FileFindNextFile($fileHandle)
			If @error Then ExitLoop

			If checkTrackFile($fileName) Then $numTracks += 1
		 WEnd
		 FileClose($fileHandle)
		 If $numTracks > 999 Then
			$id = MsgBox($MB_YESNO + $MB_ICONQUESTION + $MB_DEFBUTTON2, lang("AAH_TRACKNUM_EXCEED"), lang("AAH_TRACKNUM_EXCEED_DESC"))
			If $id <> $IDYES Then Return
			$numTracks = 999
		 EndIf

		 $fileHandle = FileFindFirstFile($phWorkSpace & "\*." & $addEncodeExt)
		 While 1
			$fileName = FileFindNextFile($fileHandle)
			If @error Then ExitLoop

			If checkTrackFile($fileName) Then

				; get track number, track artist, track composer
				$trackNumber = Number($fileName)
				$trackTitle = trimTrackNum(baseName($fileName))
				If $trackArtist[$trackNumber] Then
				$artist = $trackArtist[$trackNumber]
				Else
				$artist = $trackArtist[0]
				EndIf
				If $trackComposer[$trackNumber] Then
				$composer = $trackComposer[$trackNumber]
				Else
					$composer = $trackComposer[0]
				EndIf
				If $trackYear[$trackNumber] Then
					$year = $trackYear[$trackNumber]
				Else
					$year = $trackYear[0]
				EndIf

				; replace track specific place holders
				Local $trackInfo[9][2] = [ _
					["%tracknr%", StringFormat("%.2d", $trackNumber)], _
					["%tracknr1%", String($trackNumber)], _
					["%tracknr2%", StringFormat("%.2d", $trackNumber)], _
					["%tracknr3%", StringFormat("%.3d", $trackNumber)], _
					["%numtracks%", String($numTracks)], _
					["%year%", $year], _
					["%title%", $trackTitle], _
					["%composer%", $composer], _
					["%artist%", $artist] _
				]
				If $debug Then  _ArrayDisplay($trackInfo)

				; filter options, then call meta exe for each file (equivalent to EAC encoder call)
				$phAddMetaOptTrack = ReplacePlaceHolders($addMetaOptTrackTemplate, $trackInfo, $addMetaOptRegex)
				Call($debugger, 'Meta exe: ' & $phAddMetaExe & " " & $phAddMetaOptTrack & ' "' &  $phWorkSpace & "\" & $fileName & '"' & " (" & $addMetaOptTrack & ")")
				RunWait($phAddMetaExe & " " & $phAddMetaOptTrack & ' "' &  $phWorkSpace & "\" & $fileName & '"', "", @SW_HIDE)
				progress($trackNumber/$numTracks*100, $trackNumber, $numTracks)
			EndIf
		 WEnd
	  EndIf
	  FileClose($fileHandle)
   EndIf

	progress(100, $numTracks, $numTracks)

EndFunc

; get album art file and save it in $workSpace
; ph<var> is <var> with placeholders replaced
Func getAlbumArt()

   getFormFields()

   Local $pid, $fileHandle, $fileCopy = 0
   Local $phAlbumArtPath = replacePlaceHolders($albumArtPath, $albumInfo, "filter")
   Local $phAlbumArtFile = replacePlaceHolders($albumArtFile, $albumInfo, "filter")
   Local $phAlbumArtExe = replacePlaceHolders($albumArtExe, $albumInfo, "filter")
   Local $phAlbumArtOpt = replacePlaceHolders($albumArtOpt, $albumInfo, "filter")
   Local $phWorkSpace = replacePlaceHolders($workSpace, $albumInfo, "filter")

   ; consistency checks
   If Not checkExists($phWorkSpace) Then Return
   If Not checkExists($phAlbumArtExe) Then Return

   ; make sure any jpg file name is renamed to $albumArtFile
   FileMove($phWorkSpace & "\*.jpg", $phWorkSpace & "\" & $phAlbumArtFile, $FC_OVERWRITE)
   $fileHandle = FileFindFirstFile($phWorkSpace & "\" & $phAlbumArtFile)

   ; if file in $workSpace does not exist and album art file exists in $albumArtPath, copy it
   If $fileHandle == -1 And $albumArtPath <> "" Then
		If FileExists($phAlbumArtPath & "\" & $phAlbumArtFile) Then
			$fileCopy = FileCopy($phAlbumArtPath & "\" & $phAlbumArtFile, $phWorkSpace & "\", $FC_OVERWRITE)
		ElseIf FileExists($phAlbumArtPath & "\*.jpg") Then
			$fileCopy = FileCopy($phAlbumArtPath & "\*.jpg", $phWorkSpace & "\", $FC_OVERWRITE)
			FileMove($phWorkSpace & "\*.jpg", $phWorkSpace & "\" & $phAlbumArtFile, $FC_OVERWRITE)
		EndIf
	EndIf
   FileClose($fileHandle)

   Call($debugger, 'AlbumArtExe: ' & $phAlbumArtExe & " " & $phAlbumArtOpt & " (" & $albumArtOpt & ")")
   $pid = Run($phAlbumArtExe & ' ' & $phAlbumArtOpt)
   ; do not wait for termination, album art downloader needs more than 20 seconds to terminate
EndFunc

; add album art, replay gain, and comment, then move files to destination folder
; ph<var> is <var> with placeholders replaced
Func tagAndFile()

   getFormFields()

   Local $id, $fileHandle, $fileCopy = 0
   Local $phImageMagickExe = replacePlaceHolders($imageMagickExe, $albumInfo, "filter")
   Local $phImageMagickOpt = replacePlaceHolders($imageMagickOpt, $albumInfo, "filter")
   Local $phAlbumArtPath = replacePlaceHolders($albumArtPath, $albumInfo, "filter")
   Local $phAlbumArtFile = replacePlaceHolders($albumArtFile, $albumInfo, "filter")
   Local $phWorkSpace = replacePlaceHolders($workSpace, $albumInfo, "filter")
   Local $phEncodePath = replacePlaceHolders($encodePath, $albumInfo, "filter")
   Local $phAddEncodePath = replacePlaceHolders($addEncodePath, $albumInfo, "filter")
   Local $phMetaExe = replacePlaceHolders($metaExe, $albumInfo, "filter")
   Local $phMetaOpt = replacePlaceHolders($metaOpt, $albumInfo, "filter")
   Local $phAddMetaExe = replacePlaceHolders($addMetaExe, $albumInfo, "filter")
   Local $phAddMetaOpt = replacePlaceHolders($addMetaOpt, $albumInfo, "filter")
   Local $tracksReady

   ; consistency checks
   If Not checkExists($phWorkSpace) Then Return
   If Not checkExists($phMetaExe) Then Return
   If Not checkExists($phAddMetaExe) Then Return
   If Not checkExists($phImageMagickExe) Then Return

   ; must have at least one encoder enabled in AAH ini
   If Not $encodeEnabled And Not $addEncodeEnabled Then
	  $id = MsgBox($MB_OK + $MB_ICONERROR, lang("AAH_NOENCODER_TITLE"), lang("AAH_NOENCODER_DESC"))
	  Return
   EndIf

   ; check multi CD and get confirmation to proceed
   If $albumTotalCDs > 1 Then
		$id = MsgBox($MB_YESNO + $MB_ICONQUESTION + $MB_DEFBUTTON2, lang("AAH_MULTICD_TITLE"), lang("AAH_MULTICD_DESC"))
		If $id <> $IDYES Then Return
   EndIf

  ; make sure any jpg file name is renamed to $albumArtFile
   FileMove($phWorkSpace & "\*.jpg", $phWorkSpace & "\" & $phAlbumArtFile, $FC_OVERWRITE)
   $fileHandle = FileFindFirstFile($phWorkSpace & "\" & $phAlbumArtFile)

   If $fileHandle == -1 Then

	  ; if file in $workSpace does not exist and album art file exists in $albumArtPath, copy it
	  If $albumArtPath <> "" Then
		If FileExists($phAlbumArtPath & "\" & $phAlbumArtFile) Then
			$fileCopy = FileCopy($phAlbumArtPath & "\" & $phAlbumArtFile, $phWorkSpace & "\", $FC_OVERWRITE)
		ElseIf FileExists($phAlbumArtPath & "\*.jpg") Then
			$fileCopy = FileCopy($phAlbumArtPath & "\*.jpg", $phWorkSpace & "\", $FC_OVERWRITE)
			FileMove($phWorkSpace & "\*.jpg", $phWorkSpace & "\" & $phAlbumArtFile, $FC_OVERWRITE)
		EndIf
	  EndIf

	  If Not $fileCopy Then

		 ; file does not exist
		 $id = MsgBox($MB_YESNO + $MB_ICONQUESTION + $MB_DEFBUTTON2, lang("AAH_NOART_TITLE"), lang("AAH_NOART_DESC"))
		 If $id <> $IDYES Then Return
	  EndIf
   Else

	  ; file exists, resize image
	  If $imageMagickExe <> "" And $imageMagickOpt <> "" Then
		 Call($debugger, 'ImageMagickExe: ' & $phImageMagickExe & " " & $phImageMagickOpt & " (" & $imageMagickOpt & ")")
		 RunWait($phImageMagickExe & " " &  $phImageMagickOpt, "", @SW_HIDE)
	  EndIf

	  ; copy image to $albumArtPath
	  If $albumArtPath <> "" Then FileCopy($phWorkSpace & "\" & $phAlbumArtFile, $phAlbumArtPath & "\", $FC_OVERWRITE + $FC_CREATEPATH)
   EndIf

   FileClose($fileHandle)

   ; tag and move files
   If $encodeEnabled Then
	  $tracksReady = checkTracksReady($phWorkSpace, $encodeExt)
	  If $tracksReady>0 Then
		trimDots($phWorkSpace, $encodeExt)
		Call($debugger, 'Meta exe: ' & $phMetaExe & " " & $phMetaOpt & " (" & $metaOpt & ")")
		RunWait($phMetaExe & " " & $phMetaOpt)
		Call($debugger, 'move ' & $phWorkSpace & "\*." & $encodeExt & ' ' & $phEncodePath & "\")
		FileMove($phWorkSpace & "\*." & $encodeExt, $phEncodePath & "\", $FC_OVERWRITE + $FC_CREATEPATH)
		FileCopy($phWorkSpace & "\*.txt", $phEncodePath & "\", $FC_OVERWRITE)
		FileCopy($phWorkSpace & "\" & $phAlbumArtFile, $phEncodePath & "\", $FC_OVERWRITE)
	  ElseIf $tracksReady==0 Then
		$id = MsgBox($MB_OK + $MB_ICONERROR, lang("AAH_MUST_TAG_TITLE"), lang("AAH_MUST_TAG_DESC"))
		Return
	  EndIf
	EndIf

   If $addEncodeEnabled Then
	  $tracksReady = checkTracksReady($phWorkSpace, $addEncodeExt)
	  If $tracksReady Then
		trimDots($phWorkSpace, $encodeExt)
		Call($debugger, 'Meta exe: ' & $phAddMetaExe & " " & $phAddMetaOpt & " (" & $addMetaOpt & ")")
		RunWait($phAddMetaExe & " " & $phAddMetaOpt)
		Call($debugger, 'move ' & $phWorkSpace & "\*." & $addEncodeExt & ' ' & $phAddEncodePath & "\")
		FileMove($phWorkSpace & "\*." & $addEncodeExt, $phAddEncodePath & "\", $FC_OVERWRITE + $FC_CREATEPATH)
		FileCopy($phWorkSpace & "\*.txt", $phAddEncodePath & "\", $FC_OVERWRITE)
		FileCopy($phWorkSpace & "\" & $phAlbumArtFile, $phAddEncodePath & "\", $FC_OVERWRITE)
	  ElseIf $tracksReady==0 Then
		$id = MsgBox($MB_OK + $MB_ICONERROR, lang("AAH_MUST_TAG_TITLE"), lang("AAH_MUST_TAG_DESC"))
		Return
	  EndIf
   EndIf

   ; clean up
   FileDelete($phWorkSpace & "\*.txt")
   FileDelete($phWorkSpace & "\" & $phAlbumArtFile)

   MsgBox($MB_ICONINFORMATION, "", lang("AAH_DONE_TAG"), 5)

EndFunc

; rename files for multi CD sets
; ph<var> is <var> with placeholders replaced
Func multiCdFileRename()

   getFormFields()

   Local $phWorkSpace = replacePlaceHolders($workSpace, $albumInfo, "filter")
   Local $fileName, $fileHandle, $tracksReady
   Local $format = "%d"

   ; consistency checks
   If Not checkExists($phWorkSpace) Then Return

   ; check if it is a multi CD
   If $albumTotalCDs == 1 Then
	  MsgBox($MB_OK + $MB_ICONWARNING, lang("AAH_MULTICD_MEMBER_ERR"), lang("AAH_MULTICD_NOMEMBER"))
	  Return
   EndIf

	; format of preceding CD number digit (01-01 first track of first CD.mp3 for MultiCD sets with more than 9 CDs)
   If $albumTotalCDs >= 10 Then $format = "%02d"

   ; rename encoded files
   If $encodeEnabled Then
	  $tracksReady = checkTracksReady($phWorkSpace, $encodeExt)
	  $fileHandle = FileFindFirstFile($phWorkSpace & "\*." & $encodeExt)
	  If $fileHandle <> -1 Then
		 If $tracksReady>0 Then
			While 1
			   $fileName = FileFindNextFile($fileHandle)
			   If @error Then ExitLoop

			   ; rename file (catch files which start with "01 " or "001 " only)
			   If checkTrackFile($fileName) Then
				  If Not FileMove($phWorkSpace & "\" & $fileName, $phWorkSpace & "\" & StringFormat($format, $albumCDnumber) & "-" & $fileName) Then
					 If Not MsgBox($MB_OKCANCEL + $MB_ICONWARNING, $phWorkSpace & "\" & $fileName, lang("AAH_MULTICD_RENAME_ERR")) == $IDOK Then Return
				  EndIf
				  Call($debugger, "File " & $fileName & " renamed")
			   EndIf
			WEnd
		 ElseIf $tracksReady==0 Then
			$id = MsgBox($MB_OK + $MB_ICONERROR, lang("AAH_MUST_TAG_TITLE"), lang("AAH_MUST_TAG_DESC"))
			Return
		 EndIf
	  EndIf
	  FileClose($fileHandle)
   EndIf

   ; rename additional encoded files
   If $addEncodeEnabled Then
	  $tracksReady = checkTracksReady($phWorkSpace, $addEncodeExt)
	  $fileHandle = FileFindFirstFile($phWorkSpace & "\*." & $addEncodeExt)
	  If $fileHandle <> -1 Then
		 If $tracksReady>0 Then
			While 1
			   $fileName = FileFindNextFile($fileHandle)
			   If @error Then ExitLoop

			   ; rename file (catch files which start with "01 " or "001 " only)
			   If checkTrackFile($fileName) Then
				  If Not FileMove($phWorkSpace & "\" & $fileName, $phWorkSpace & "\" & StringFormat($format, $albumCDnumber) & "-" & $fileName) Then
					 If Not MsgBox($MB_OKCANCEL + $MB_ICONWARNING, $phWorkSpace & "\" & $fileName, lang("AAH_MULTICD_RENAME_ERR")) == $IDOK Then Return
				  EndIf
				  Call($debugger, "File " & $fileName & " renamed")
			   EndIf
			WEnd
		 ElseIf $tracksReady==0 Then
			$id = MsgBox($MB_OK + $MB_ICONERROR, lang("AAH_MUST_TAG_TITLE"), lang("AAH_MUST_TAG_DESC"))
			Return
		 EndIf
	  EndIf
	  FileClose($fileHandle)
   EndIf

   ; find track artists files and rename them
   If $encodeEnabled Or $addEncodeEnabled Then
	  $fileHandle = FileFindFirstFile($phWorkSpace & "\*.txt")
	  If $fileHandle <> -1 Then
		 While 1
			$fileName = FileFindNextFile($fileHandle)
			If @error Then ExitLoop

			; rename and add to table (catch files which start with a non-digit character only)
			If checkInfoFile($fileName) Then
			   If Not FileMove($phWorkSpace & "\" & $fileName, $phWorkSpace & "\" & StringFormat($format, $albumCDnumber) & "-" & $fileName) Then
				  If Not MsgBox($MB_OKCANCEL + $MB_ICONWARNING, $phWorkSpace & "\" & $fileName, lang("AAH_MULTICD_RENAME_ERR")) == $IDOK Then Return
			   EndIf
			   Call($debugger, "File " & $fileName & " renamed")
			EndIf
		 WEnd
	  EndIf
	  FileClose($fileHandle)
   EndIf

   MsgBox($MB_ICONINFORMATION, "", lang("AAH_DONE_MULTI"), 5)

EndFunc

; set 'Various Artists' as $albumartist and mark CD as compilation
Func setVariousArtists()

   getFormFields()

   Local $phWorkSpace = replacePlaceHolders($workSpace, $albumInfo, "filter")
   Local $phNotepadExe = replacePlaceHolders($notepadExe, $albumInfo, "filter")
   Local $notepadTitle = lang("AAH_NOTEPAD_TITLE")

   ; consistency checks
   If Not checkExists($phWorkSpace) Then Return

   $vaSet = 1
   $metaArtist = GUICtrlRead($formFieldArtist)
   ControlSetText($windowTitle, "", $formFieldArtist, $vaName, 1)
   Call($debugger, lang("AAH_MENU_SET_VA"))

   ; we need to create a track artist file
   If WinExists($trackArtistsFile & " - " & $notepadTitle) Then
	  MsgBox($MB_OK + $MB_ICONWARNING, lang("AAH_FILE_TITLE") & ": " & $trackArtistsFile, lang("AAH_NOTEPAD_OPEN"))
	  Return
   EndIf
   $sendHelp = Not FileExists($phWorkSpace & "\" & $trackArtistsFile)
   Call($debugger, $phNotepadExe & ' "' & $phWorkSpace & '\' & $trackArtistsFile & '" - ' & $notepadTitle)
   Run($phNotepadExe & ' "' & $phWorkSpace & '\' & $trackArtistsFile & '"')

   If $sendHelp Then

	  ; create a new file, then send help lines to the file
	  WinWaitActive($notepadTitle)
	  Send("{ENTER}")
	  WinWaitActive($trackArtistsFile & " - " & $notepadTitle)
	  Send(lang("AAH_ARTIST_TRACK1") & "{ENTER}" & lang("AAH_ARTIST_TRACK2") & "{ENTER}" & lang("AAH_ARTIST_TRACK3") & "{ENTER}" & lang("AAH_ARTIST_TRACK4") & "{ENTER}" & lang("AAH_ARTIST_TRACK5") & "{ENTER}")
	  Sleep(100)

	  ; wait until the file has been saved, then close notepad
	  ; this works because the window title is preceded by a '*' after having sent the text
	  Do
		 Sleep(100)
	  Until WinExists($trackArtistsFile & " - " & $notepadTitle) Or Not WinExists("*" & $trackArtistsFile & " - " & $notepadTitle)
	  WinClose($trackArtistsFile & " - " & $notepadTitle)
   EndIf
EndFunc

Func removeVariousArtists()

   getFormFields()

   Local $phWorkSpace = replacePlaceHolders($workSpace, $albumInfo, "filter")

   ; consistency checks
   If Not checkExists($phWorkSpace) Then Return

   $vaSet = 0
   GUICtrlSetData($formFieldArtist, $metaArtist)
   Call($debugger, lang("AAH_MENU_REM_VA"))

   If FileExists($workSpace & "\" & $trackArtistsFile) Then
	  $id = MsgBox($MB_YESNO + $MB_ICONQUESTION + $MB_DEFBUTTON2, lang("AAH_FILE_TITLE") & ": " & $trackArtistsFile, lang("AAH_REMOVE_FILE"))
	  If $id == $IDYES Then FileDelete($phWorkSpace & "\" & $trackArtistsFile)
   EndIf
EndFunc

; main
; ----------------------------------------------------------------------------------------------------
initialize()

; Exit if AAC is already running
If WinExists($windowTitle) Then
   MsgBox($MB_OK + $MB_ICONWARNING, lang("AAH_RUNNING_ERR"), lang("AAH_RUNNING_DESC"))
   Exit
EndIf

guiMain()
terminate()

Exit

; main program GUI
; ----------------------------------------------------------------------------------------------------
Func guiMain()

	; create window
   $windowHandle = GUICreate($windowTitle, 800, 200)
   insertMenu()
   createFormFields()

   GUIRegisterMsg(0x0053, "WM_HELP")
   GUISetState()

   While 1
	  $cmd = GUIGetMsg()
	  If $cmd > 0 Then Call($debugger, "$cmd=" & $cmd)
	  If $cmd == $GUI_EVENT_CLOSE Or $cmd == $command[UBound($command, $UBOUND_ROWS) - 1][1] Then Exit
	  For $i = 0 To UBound($command, $UBOUND_ROWS) - 2
		 If $cmd == $command[$i][1] Then
			Call($debugger, "command=" & $command[$i][2])
			Call($command[$i][2])
		 EndIf
	  Next
	  If $cmd == $formFieldVA Then
		 If GUICtrlRead($formFieldVA) == $GUI_CHECKED  Then
			setVariousArtists()
		 Else
			removeVariousArtists()
		 EndIf
	  EndIf
	  Sleep(50)
   WEnd
EndFunc
