﻿#cs
EAH - Exact Audio Helper
AutoIt v3.3.14.5 - http://www.autoitscript.com

Copyright (c) 2020, Bernard Condrau.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------------------------------
This script was inspired by REACT (c) Tycho, 2006
- see http://www.hydrogenaudio.org/forums/index.php?showtopic=35158
- Exact Audio Helper serves a similar purpose as REACT, but has different functionality

Exact Audio Helper was compiled with AutoIt v3.3.14.5 for Exact Audio Copy v1.5
- see available place holders at the end of this file
----------------------------------------------------------------------------------------------------
The default settings for metaflac and metamp3 complement the following EAC encoder options
EAC
mp3 (lame): -V4 --vbr-new --id3v2-only --replaygain-accurate --ignore-tag-errors --ta "%artist%" --tt "%title%" --tl "%albumtitle%" --ty "%year%" --tn "%tracknr1%/%numtracks%" --tg "%genre%" --tv "TPE2=%albumartist%" --tv "TPE3=%albuminterpret%" --tv "TPOS=%cdnumber%/%totalcds%" --tv "TCOM=%composer%" %source% %dest%
flac: -8 -V -T "ARTIST=%artist%" -T "TITLE=%title%" -T "ALBUM=%albumtitle%" -T "DATE=%year%" -T "TRACKNUMBER=%tracknr1%" -T "TOTALTRACKS=%numtracks%" -T "GENRE=%genre%" -T "ALBUMARTIST=%albumartist%" -T "PERFORMER=%albuminterpret%" -T "DISCNUMBER=%cdnumber%" -T "TOTALDISCS=%totalcds%" -T "COMPOSER=%composer%" %source% -o %dest%
EAH
metamp3 : --2 --replay-gain --frame "APIC{3}:%workspace%\%coverfile%" --comment "%comment%" --2 "%workspace%\*.mp3"
metaflac: --add-replay-gain --import-picture-from="%workspace%\%coverfile%" --set-tag=COMMENT="%comment%" --show-tag=TITLE "%workspace%\*.flac"
----------------------------------------------------------------------------------------------------
Note on debugging this script
- Set Globals $debug = 1 and $breakPointExe
- #include <Debug.au3>
- $breakPointExe points to a local DOS command batch file on your computer with following content
  @ECHO OFF
  ECHO Break Point: %1
  PAUSE
- Whenever you call function breakPoint(), the script pauses until hit a key in the DOS batch File
  This helps trouble shoot file modification, copy, and move commands
----------------------------------------------------------------------------------------------------
#ce

#pragma compile(UPX, True)
#pragma compile(AutoItExecuteAllowed, False)
#pragma compile(FileVersion, 1.0.20.1130)
#pragma compile(ProductName, Exact Audio Helper)
#pragma compile(ProductVersion, 1.0)
#pragma compile(LegalCopyright, (c) 2020 Bernard Condrau)
#pragma compile(Comments, GNU General Public License version 3)
#pragma compile(OriginalFilename, EAH.exe)

#NoTrayIcon

#include <WindowsConstants.au3>
#include <MsgBoxConstants.au3>
#include <GuiMenu.au3>
#include <GuiButton.au3>
#include <Array.au3>
#include <File.au3>

; set debug=0 for final build
Global $debug = 0
Global $breakPointExe = '"C:\PortableCustom\EAH\tools\breakpoint.bat"'
;#include <Debug.au3>

; app
; ----------------------------------------------------------------------------------------------------
Global $appName    = "Exact Audio Helper"
Global $appVersion = "1.0.20.1130" ; same as #pragma above
Global $appDate    = "30 Nov 2020"
Global $appHelpURL = "https://wiki.condrau.com/media:eah"
Global $appAbout   = " EAH..."

; form fields
; ----------------------------------------------------------------------------------------------------
Global $formFieldArtist		= "myedit3" ; %albumartist%
Global $formFieldAlbum		= "myedit2" ; %albumtitle%
Global $formFieldComposer	= "myedit6" ; %albumcomposer%
Global $formFieldPerformer	= "myedit4" ; %albuminterpret%
Global $formFieldCDnumber	= "myedit8" ; %cdnumber%
Global $formFieldTotalCDs	= "myedit9" ; %totalcds%
Global $formFieldGenre		= "Edit1"	; %genre%
Global $formFieldYear		= "myedit5"	; %year%
Global $formFieldComment	= "myedit7" ; %comment%

; globals
; ----------------------------------------------------------------------------------------------------
Global $scriptName = StringTrimRight(@ScriptName, 4)
Global $iniFile = @ScriptDir & "\" & $scriptName & ".ini"
Global $langFile = @ScriptDir & "\" & $scriptName & ".en.ini"
Global $logFileExt = "log" ; log file extension
Global $menuCnt, $menuHandle
Global $windowHandle, $menuHandle
Global $windowTitle = "Exact Audio Copy"
Global $windowEahTitle = $appName
Global $eacExeRegistry = RegRead("HKEY_CURRENT_USER\Software\AWSoftware\EACU", "InstallPath") & "EAC.exe"
Global $myMusicRegistry = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\User Shell Folders", "My Music")
Global $albumArtExeRegistry = RegRead("HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\AlbumArt.exe", "")
Global $albumArtOptPreset = '"%albumartist%" "%albumtitle%" ""%workspace%\%coverfile%"" /group local /autoclose'
Global $imageMagickExeRegistry = RegRead("HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\ImageMagick.exe", "")
Global $imageMagickOptPreset = '-depth 8 "%workspace%\%coverfile%" "%workspace%\%coverfile%"'
Global $eahCommentDefault = "Created with EAC/" & $scriptName & ", %today%"
Global $vaNameDefault = "Various Artists"
Global $language,$eacExe,$workSpace,$eahComment,$vaName,$registerHotKeys
Global $metaExe,$metaOpt,$encodeExt,$encodePath,$encodeEnabled,$addMetaExe,$addMetaOpt,$addEncodeExt,$addEncodePath,$addEncodeEnabled
Global $albumArtPath,$albumArtExe,$albumArtOpt,$albumArtFile,$imageMagickExe,$imageMagickOpt,$saveCoverImage,$resizeCoverImage

; debugging
; ----------------------------------------------------------------------------------------------------
If $debug Then
	; avoid compile error missing function _DebugSetup()
	$debugger = "_DebugSetup"
	Call($debugger)
	$debugger = "_DebugReport"
	Call($debugger, "AutoIt runs x64 = " & @AutoItX64)
Else
	$debugger = "noDebugReport"
EndIf

Func noDebugReport($sData, $bLastError = False, $bExit = False)
   ; this function replaces _DebugReport when debugging is off, and does nothing
EndFunc

Func breakPoint($position = @scriptName)
   If $debug Then RunWait($breakPointExe & " " & $position)
EndFunc

; from UDF Misc.au3, avoid loading the complete UDF for this one function
; ----------------------------------------------------------------------------------------------------
Func _IsPressed($sHexKey, $vDLL = "user32.dll")
	Local $aReturn = DllCall($vDLL, "short", "GetAsyncKeyState", "int", "0x" & $sHexKey)
	If @error Then Return SetError(@error, @extended, False)
	Return BitAND($aReturn[0], 0x8000) <> 0
EndFunc   ;==>_IsPressed

; helpers
; ----------------------------------------------------------------------------------------------------
; callback for F1 and Help button
Func WM_HELP()
	ShellExecute($appHelpURL)
	Return
EndFunc

Func registerHotKey($cmd, $text = 0)
   If $registerHotKeys Then
	  Local $hotKey[4][3] = [ _
		 ["eacCopyTracks", "", @TAB & "Shift-F5"], _ ; EAC built-in hotkey
		 ["getAlbumArt", "+{F9}", @TAB & "Shift-F9"], _
		 ["tagAndFile", "^{F9}", @TAB & "Ctrl-F9"], _
		 ["help", "", @TAB & "F1"] _ ; F1 is the predefined help key
	  ]
	  For $i = 0 To UBound($hotKey, $UBOUND_ROWS) - 1
		 If $hotKey[$i][0] == $cmd Then Return $hotKey[$i][1+$text]
	  Next
   EndIf
   Return ""
EndFunc

Func insertMenu()

   Local $itemHandle, $itemText, $newItemText, $accessKey, $accessKeyPos

   ; set menu text and command IDs
   Global $command[7][3] = [ _
	  [lang("EAH_MENU_COPY"), 0x2000, "eacCopyTracks"], _
	  [lang("EAH_MENU_ART"), 0x2100, "getAlbumArt"], _
	  [lang("EAH_MENU_TAG"), 0x2200, "tagAndFile"], _
	  [lang("EAH_MENU_MULTI"), 0x2300, "multiCdFileRename"], _
	  ["", -1], _
	  [lang("EAH_MENU_ABOUT") & $appAbout, 0x2400, "about"], _
	  [lang("EAH_MENU_HELP"), 0x2500, "help"] _
   ]

   Call($debugger, "Inserting menu")

    $menuHandle = _GUICtrlMenu_GetMenu($windowHandle)
	$menuCnt = _GUICtrlMenu_GetItemCount($menuHandle)

   ; create menu and hotkeys
   $itemHandle = _GUICtrlMenu_CreateMenu()
   For $i = 0 To UBound($command, $UBOUND_ROWS) - 1
	  $command[$i][0] = $command[$i][0] & registerHotKey($command[$i][2], 1)
	  _GUICtrlMenu_InsertMenuItem($itemHandle, $i, $command[$i][0], $command[$i][1])
	  HotKeySet(registerHotKey($command[$i][2]), $command[$i][2])
   Next

   ; override access key in menu to our 'Encode'
   $newItemText = lang("EAH_MENU_ENCODE")
   $accessKeyPos = StringInStr($newItemText, "&")
   If $accessKeyPos > 0 Then
	  $accessKey = StringMid($newItemText, $accessKeyPos, 2)
	  For $n = 0 To $menuCnt - 1
		 $itemText = _GUICtrlMenu_GetItemText($menuHandle, $n)
		 If StringInStr($itemText, $accessKey) Then _GUICtrlMenu_SetItemText ($menuHandle, $n, StringTrimLeft($itemText, 1))
	  Next
   EndIf

   ; insert new menu item
   _GUICtrlMenu_InsertMenuItem($menuHandle, $menuCnt, $newItemText, 0, $itemHandle)
   _GUICtrlMenu_DrawMenuBar($windowHandle)

   Call($debugger, "Inserting menu done")
EndFunc

; create additional form fields in window
Func createFormFields()

   Global $vaHandle, $vaState

   ; various artists checkbox
   $vaHandle = _GUICtrlButton_Create($windowHandle, lang("EAH_MENU_SET_VA"), 432, 75, 164, 18, $BS_AUTOCHECKBOX)
   $vaState = 0
EndFunc

Func getFormFields()

   Call($debugger, "Get form field values")

   Global $albumCDnumber, $albumTotalCDs ; we need these to handle CD numbering
   Local $albumArtist, $albumTitle, $albumComposer
   Local $albumPerformer, $albumComment, $albumYear, $albumGenre, $today

   ; get form fields
   $albumTitle = ControlGetText($windowTitle, "", $formFieldAlbum)
   $albumArtist = ControlGetText($windowTitle, "", $formFieldArtist)
   $albumComposer = ControlGetText($windowTitle, "", $formFieldComposer)
   $albumPerformer = ControlGetText($windowTitle, "", $formFieldPerformer)
   $albumComment = ControlGetText($windowTitle, "", $formFieldComment)
   $albumYear = ControlGetText($windowTitle, "", $formFieldYear)
   $albumGenre = ControlGetText($windowTitle, "", $formFieldGenre)
   $albumCDnumber = ControlGetText($windowTitle, "", $formFieldCDnumber)
   $albumTotalCDs = ControlGetText($windowTitle, "", $formFieldTotalCDs)
   $today = @YEAR & "-" & @MON & "-" & @MDAY ; must be file system conform

   ; we need to combine comments from the "Comment" form field and ini file
   If $albumComment == "" Then
	  $albumComment = $eahComment
   EndIf

   Global $albumInfo[14][3] = [ _
	  ["%coverfile%", $albumArtFile, replaceCharacters($albumArtFile)], _ ; must be first place holder as this might contain other place holders
	  ["%albumtitle%", $albumTitle, replaceCharacters($albumTitle)], _
	  ["%albumartist%", $albumArtist, replaceCharacters($albumArtist)], _
	  ["%albumcomposer%", $albumComposer, replaceCharacters($albumComposer)], _
	  ["%albuminterpret%", $albumPerformer, replaceCharacters($albumPerformer)], _
	  ["%comment%", $albumComment, $albumComment], _
	  ["%year%", $albumYear, $albumYear], _
	  ["%genre%", $albumGenre, $albumGenre], _
	  ["%cdnumber%", $albumCDnumber, $albumCDnumber], _
	  ["%totalcds%", $albumTotalCDs, $albumTotalCDs], _
	  ["%workspace%", $workSpace, $workSpace], _
	  ["%scriptdir%", @ScriptDir, @ScriptDir], _
	  ["%username%", @UserName, @UserName], _
	  ["%today%", $today, $today] _
   ]
   If $debug Then _ArrayDisplay($albumInfo)
EndFunc

Func iniFileRead($filename, $section, $key, $default, $exe = 0)

   Local $value

   $value = IniRead($filename, $section, $key, $default)
   If $value == "" Then $value = $default
   If $exe Then quotePath($value)
   Return $value

EndFunc

Func initialize()

	; since we prepend EAH to the title we want to find the window at all times
	Opt("WinTitleMatchMode", 2)
	Global $dll = DllOpen("user32.dll")

	Local $myLanguage

	Call($debugger, "initialize")

	; check whether my language translation file exists
	$myLanguage = IniFileRead($iniFile, "Settings", "Language", "en")
	If FileExists(@ScriptDir & "\" & $scriptName & "." & $myLanguage & ".ini") Then
		$language = $myLanguage
		$langFile = @ScriptDir & "\" & $scriptName & "." & $language & ".ini"
	EndIf

	; read ini file - Settings
	$eacExe = IniFileRead($iniFile, "Settings", "EAC_Exe", $eacExeRegistry, 1)
	$workSpace = IniFileRead($iniFile, "Settings", "Workspace", $myMusicRegistry)
	$eahComment = IniFileRead($iniFile, "Settings", "Comment", $eahCommentDefault)
	$vaName = IniFileRead($iniFile, "Settings", "VA", $vaNameDefault)
	$registerHotKeys = Int(IniFileRead($iniFile, "Settings", "RegisterHotKeys", 0))

	; read ini file - Encoders
	$metaExe = IniFileRead($iniFile, "Encoders", "MetaExe", "", 1)
	$metaOpt = IniFileRead($iniFile, "Encoders", "MetaOpt", "")
	$encodeExt = IniFileRead($iniFile, "Encoders", "EncodeExt", "")
	$encodePath = IniFileRead($iniFile, "Encoders", "EncodePath", "")
	$addMetaExe = IniFileRead($iniFile, "Encoders", "AddMetaExe", "", 1)
	$addMetaOpt = IniFileRead($iniFile, "Encoders", "AddMetaOpt", "")
	$addEncodeExt = IniFileRead($iniFile, "Encoders", "AddEncodeExt", "")
	$addEncodePath = IniFileRead($iniFile, "Encoders", "AddEncodePath", "")
	$encodeEnabled = $metaExe <> "" And $metaOpt <> "" And $encodeExt <> "" And $encodePath <> ""
	$addEncodeEnabled = $addMetaExe <> "" And $addMetaOpt <> "" And $addEncodeExt <> "" And $addEncodePath <> ""

	; read ini file - AlbumArt
	$albumArtFile = IniFileRead($iniFile, "AlbumArt", "AlbumArtFile", "folder.jpg")
	$albumArtExe = IniFileRead($iniFile, "AlbumArt", "AlbumArtExe", $albumArtExeRegistry, 1)
	$albumArtOpt = IniFileRead($iniFile, "AlbumArt", "AlbumArtOpt", $albumArtOptPreset)
	$albumArtPath = IniFileRead($iniFile, "AlbumArt", "AlbumArtPath", "")
	$imageMagickExe = IniFileRead($iniFile, "AlbumArt", "ImageMagickExe", $imageMagickExeRegistry, 1)
	$imageMagickOpt = IniFileRead($iniFile, "AlbumArt", "ImageMagickOpt", $imageMagickOptPreset)
	$saveCoverImage = $albumArtPath <> ""
	$resizeCoverImage = $imageMagickExe <> "" And $imageMagickOpt <> ""

	; read character replacements table from registry and/or ini file
	getCharacterReplacements()

	; remove trailing \
	trimSlash($workSpace)
	trimSlash($encodePath)
	trimSlash($addEncodePath)
	trimSlash($albumArtPath)

	Call($debugger, 'Reading ' & $iniFile & '...' & @CRLF & $language & ' (' & lang("EAH_LANGUAGE") & ') - ' & $langFile & @CRLF & $eacExe & @CRLF & $albumArtExe & @CRLF & $albumArtFile & @CRLF & $workSpace & @CRLF & $encodePath & @CRLF & $addEncodePath)
	Call($debugger, $albumArtPath & @CRLF & $eahComment & @CRLF & $vaName & @CRLF & "Register hotkeys=" & $registerHotKeys)

EndFunc

Func lang($key)
   $str = IniRead($langFile, $language, $key, $key)
   Return StringReplace(StringReplace($str, "%TAB%", @TAB), "%CRLF%", @CRLF)
EndFunc

Func terminate()
   Call($debugger, 'terminate')
   DllClose($dll)
EndFunc

; file name manipulation
; ----------------------------------------------------------------------------------------------------
Func checkExists($path)
	Local $langMsg
	$path = StringRegExpReplace($path, '^"|"$', "")
	If $path <> "" And Not FileExists($path) Then
		$langMsg = "EAH_PATH_NOT_FOUND"
		If StringRegExp($path, '.*\.exe$') Then $langMsg = "EAH_FILE_NOT_FOUND"
		MsgBox($MB_OK + $MB_ICONWARNING, $path, lang($langMsg))
		Return False
	EndIf
	Return True
EndFunc

; verify and rename valid file names
; check whether audacity file names have been renamed
Global Const $regexProcessPattern = '^([\d]{2,3})\s([^\s])'		; start with "01 " or "001 " not followed by another space
Global Const $regexProcessReplace = '$2'						; replacement pattern (01 filename => filename, 001 filename => filename)

; check whether file name can be processed
Func checkTrackFile($str)
	Return StringRegExp($str, $regexProcessPattern)
EndFunc

; verify info file
Global Const $regexInfoPattern = '^([\d]{1,2})-([^\s])'			; start with "1-" or "01-" not followed by space

; check whether info file can be processed
Func checkInfoFile($str)
	Return Not StringRegExp($str, $regexInfoPattern)
EndFunc

; remove more than 1 dot before the extension (metamp3.exe cannot handle it)
Func trimDots($path, $ext)

	Local $filteredName, $fileName, $fileHandle

	If $ext<>"" Then
		$fileHandle = FileFindFirstFile($path & "\*." & $ext)
		While 1
			$fileName = FileFindNextFile($fileHandle)
			If @error Then ExitLoop

			$filteredName = StringStripWS(StringRegExpReplace($fileName, "[.]{2,}" & $ext, "." & $ext), $STR_STRIPLEADING + $STR_STRIPTRAILING)
			If $filteredName <> $fileName Then FileMove($path & "\" & $fileName, $path & "\" & $filteredName)
		WEnd
	EndIf
EndFunc

; read character replacements table from EAC registry or ini file
Func getCharacterReplacements()

	Global $characterTable[0][2]
	Local $slash = False
	Local $empty = 16 ; add "/" to the end of the table if no table entry is empty

	For $i = 0 To 15
		$rule = IniFileRead($iniFile, "CharacterReplacements", "ReplacementRule" & $i, RegRead("HKEY_CURRENT_USER\Software\AWSoftware\EACU\Extraction Options", "ReplacementRule" & $i), 0)
		If $rule <> "" Then
			If StringLeft($rule, 1) == "/" Then $slash = True
			_ArrayAdd($characterTable, StringLeft($rule, 1) & "/?/" & StringTrimLeft($rule, 1), 0, "/?/")
		Else
			If $empty > 15 Then $empty = $i
			_ArrayAdd($characterTable, "")
		EndIf
	Next
	If Not $slash Then
		$characterTable[$empty][0] = "/"
		$characterTable[$empty][1] = ","
	EndIf
	If $debug Then _ArrayDisplay($characterTable)
EndFunc

; remove illegal file name characters
Func replaceCharacters($str)

	; replace characters
	For $i = 0 To UBound($characterTable) - 1
		If $characterTable[$i][0] <> "" Then $str = StringReplace($str, $characterTable[$i][0], $characterTable[$i][1], 0, $STR_CASESENSE)
	Next

	; remove preceding spaces
	While StringLeft($str, 1) == ' '
		$str = StringTrimLeft($str, 1)
	Wend

	; remove trailing spaces or dots
	While StringRight($str, 1) == ' ' Or StringRight($str, 1) == '.'
		$str = StringTrimRight($str, 1)
	Wend

	Return $str
EndFunc

; trim trailing slash
Func trimSlash(ByRef $path)
   If StringRight($path, 1) == '\' Then $path = StringTrimRight($path, 1)
EndFunc

; AutoIt's Run command requires quoted path/exe names
Func quotePath(ByRef $path)
   If StringLeft($path, 1) == '"' And StringRight($path, 1) == '"' Then Return
   If $path == "" Then Return
   $path = '"' & $path & '"'
EndFunc

; replace place holders
;	$str: string with place holders
;	$table: replacement table with patterns and replacement Strings. Col0=pattern, Col1=replacement string, Col2=filtered replacement string.
;	$regex: ""=no further processing, "filter"=replace illegal (file system) characters, "<regex string>"=regex pattern to replace with an empty string. Separate several regex patterns with " ".
;	Return: filtered and replaced string
; Note: if you need to filter and remove regex patterns, call the function twice
Func replacePlaceHolders($str, Const ByRef $table, $regex = "filter")

   Local $col = 1 ; column not filtered
   If $regex=="filter" Then $col = 2 ; column filtered

   ; replace place holders
   For $i = 0 to UBound($table, $UBOUND_ROWS) - 1
	  $str = StringReplace($str, $table[$i][0], $table[$i][$col])
   Next

   ; remove regex patterns
   If $regex<>"raw" And $regex<>"filter" Then
	 $pattern = StringSplit($regex, " ")
	 For $i = 1 to $pattern[0]
		$str = StringRegExpReplace($str, $pattern[$i], "")
	 Next
   EndIf

   Return $str
EndFunc

; actions
; ----------------------------------------------------------------------------------------------------
; about msg box
Func about()
	MsgBox($MB_ICONINFORMATION, lang("EAH_MENU_ABOUT") & " " & $appName, $appName & " v" & $appVersion & " (" & $appDate & ")" & @CRLF & $appHelpURL, 10)
EndFunc

; invoke help
Func help()
	WM_HELP()
EndFunc

; call EAC's built in "Copy Selected Tracks --> Compressed..."
Func eacCopyTracks()

   Call($debugger, lang("EAH_MENU_COPY"))
   WinActivate("[CLASS:#32768]")
   Send("+{F5}")
   Sleep(50)
   WinActivate($windowTitle)
EndFunc

; get album art file and save it in $workSpace
; ph<var> is <var> with placeholders replaced
Func getAlbumArt()

   getFormFields()

   Local $pid, $fileHandle, $fileCopy = 0
   Local $phAlbumArtPath = replacePlaceHolders($albumArtPath, $albumInfo)
   Local $phAlbumArtFile = replacePlaceHolders($albumArtFile, $albumInfo)
   Local $phAlbumArtExe = replacePlaceHolders($albumArtExe, $albumInfo, "filter")
   Local $phAlbumArtOpt = replacePlaceHolders($albumArtOpt, $albumInfo)
   Local $phWorkSpace = replacePlaceHolders($workSpace, $albumInfo)

   ; consistency checks
   If Not checkExists($phWorkSpace) Then Return
   If Not checkExists($phAlbumArtExe) Then Return

   ; make sure any jpg file name is renamed to $albumArtFile
   FileMove($phWorkSpace & "\*.jpg", $phWorkSpace & "\" & $phAlbumArtFile, $FC_OVERWRITE)
   $fileHandle = FileFindFirstFile($phWorkSpace & "\" & $phAlbumArtFile)

   ; if file in $workSpace does not exist and album art file exists in $albumArtPath, copy it
   If $fileHandle == -1 And $albumArtPath <> "" Then
		If FileExists($phAlbumArtPath & "\" & $phAlbumArtFile) Then
			$fileCopy = FileCopy($phAlbumArtPath & "\" & $phAlbumArtFile, $phWorkSpace & "\", $FC_OVERWRITE)
		ElseIf FileExists($phAlbumArtPath & "\*.jpg") Then
			$fileCopy = FileCopy($phAlbumArtPath & "\*.jpg", $phWorkSpace & "\", $FC_OVERWRITE)
			FileMove($phWorkSpace & "\*.jpg", $phWorkSpace & "\" & $phAlbumArtFile, $FC_OVERWRITE)
		EndIf
	EndIf
   FileClose($fileHandle)

   Call($debugger, 'AlbumArtExe: ' & $phAlbumArtExe & " " & $phAlbumArtOpt & " (" & $albumArtOpt & ")")
   $pid = Run($phAlbumArtExe & ' ' & $phAlbumArtOpt)
   ; do not wait for termination, album art downloader needs more than 20 seconds to terminate
EndFunc

; add album art, replay gain, and comment, then move files to destination folder
; ph<var> is <var> with placeholders replaced
Func tagAndFile()

   getFormFields()

   Local $id, $fileHandle, $fileCopy = 0
   Local $phImageMagickExe = replacePlaceHolders($imageMagickExe, $albumInfo, "filter")
   Local $phImageMagickOpt = replacePlaceHolders($imageMagickOpt, $albumInfo)
   Local $phAlbumArtPath = replacePlaceHolders($albumArtPath, $albumInfo)
   Local $phAlbumArtFile = replacePlaceHolders($albumArtFile, $albumInfo)
   Local $phWorkSpace = replacePlaceHolders($workSpace, $albumInfo)
   Local $phEncodePath = replacePlaceHolders($encodePath, $albumInfo)
   Local $phAddEncodePath = replacePlaceHolders($addEncodePath, $albumInfo)
   Local $phMetaExe = replacePlaceHolders($metaExe, $albumInfo)
   Local $phMetaOpt = replacePlaceHolders($metaOpt, $albumInfo)
   Local $phAddMetaExe = replacePlaceHolders($addMetaExe, $albumInfo)
   Local $phAddMetaOpt = replacePlaceHolders($addMetaOpt, $albumInfo)

   ; consistency checks
   If Not checkExists($phWorkSpace) Then Return
   If Not checkExists($phMetaExe) Then Return
   If Not checkExists($phAddMetaExe) Then Return
   If Not checkExists($phImageMagickExe) Then Return

   ; must have at least one encoder enabled in EAH ini
   If Not $encodeEnabled And Not $addEncodeEnabled Then
	  $id = MsgBox($MB_OK + $MB_ICONERROR, lang("EAH_NOENCODER_TITLE"), lang("EAH_NOENCODER_DESC"))
	  Return
   EndIf

   ; check multi CD and get confirmation to proceed
   If $albumTotalCDs > 1 Then
		$id = MsgBox($MB_YESNO + $MB_ICONQUESTION + $MB_DEFBUTTON2, lang("EAH_MULTICD_TITLE"), lang("EAH_MULTICD_DESC"))
		If $id <> $IDYES Then Return
   EndIf

  ; make sure any jpg file name is renamed to $albumArtFile
   FileMove($phWorkSpace & "\*.jpg", $phWorkSpace & "\" & $phAlbumArtFile, $FC_OVERWRITE)
   $fileHandle = FileFindFirstFile($phWorkSpace & "\" & $phAlbumArtFile)

   If $fileHandle == -1 Then

	  ; if file in $workSpace does not exist and album art file exists in $albumArtPath, copy it
	  If $albumArtPath <> "" Then
		If FileExists($phAlbumArtPath & "\" & $phAlbumArtFile) Then
			$fileCopy = FileCopy($phAlbumArtPath & "\" & $phAlbumArtFile, $phWorkSpace & "\", $FC_OVERWRITE)
		ElseIf FileExists($phAlbumArtPath & "\*.jpg") Then
			$fileCopy = FileCopy($phAlbumArtPath & "\*.jpg", $phWorkSpace & "\", $FC_OVERWRITE)
			FileMove($phWorkSpace & "\*.jpg", $phWorkSpace & "\" & $phAlbumArtFile, $FC_OVERWRITE)
		EndIf
	  EndIf

	  If Not $fileCopy Then

		 ; file does not exist
		 $id = MsgBox($MB_YESNO + $MB_ICONQUESTION + $MB_DEFBUTTON2, lang("EAH_NOART_TITLE"), lang("EAH_NOART_DESC"))
		 If $id <> $IDYES Then Return
	  EndIf
   Else

	  ; file exists, resize image
	  If $imageMagickExe <> "" And $imageMagickOpt <> "" Then
		 Call($debugger, 'ImageMagickExe: ' & $phImageMagickExe & " " & $phImageMagickOpt & " (" & $imageMagickOpt & ")")
		 RunWait($phImageMagickExe & " " &  $phImageMagickOpt, "", @SW_HIDE)
	  EndIf

	  ; copy image to $albumArtPath
	  If $albumArtPath <> "" Then FileCopy($phWorkSpace & "\" & $phAlbumArtFile, $phAlbumArtPath & "\", $FC_OVERWRITE + $FC_CREATEPATH)
   EndIf

   FileClose($fileHandle)

	; tag and move files
	If $encodeEnabled Then
		trimDots($phWorkSpace, $encodeExt)
		Call($debugger, 'Meta exe: ' & $phMetaExe & " " & $phMetaOpt & " (" & $metaOpt & ")")
		RunWait($phMetaExe & " " & $phMetaOpt)
		Call($debugger, 'move ' & $phWorkSpace & "\*." & $encodeExt & ' ' & $phEncodePath & "\")
		FileMove($phWorkSpace & "\*." & $encodeExt, $phEncodePath & "\", $FC_OVERWRITE + $FC_CREATEPATH)
		FileCopy($phWorkSpace & "\*.log", $phEncodePath & "\", $FC_OVERWRITE)
		FileCopy($phWorkSpace & "\" & $phAlbumArtFile, $phEncodePath & "\", $FC_OVERWRITE)
	EndIf

	If $addEncodeEnabled Then
		trimDots($phWorkSpace, $addEncodeExt)
		Call($debugger, 'Meta exe: ' & $phAddMetaExe & " " & $phAddMetaOpt & " (" & $addMetaOpt & ")")
		RunWait($phAddMetaExe & " " & $phAddMetaOpt)
		Call($debugger, 'move ' & $phWorkSpace & "\*." & $addEncodeExt & ' ' & $phAddEncodePath & "\")
		FileMove($phWorkSpace & "\*." & $addEncodeExt, $phAddEncodePath & "\", $FC_OVERWRITE + $FC_CREATEPATH)
		FileCopy($phWorkSpace & "\*.log", $phAddEncodePath & "\", $FC_OVERWRITE)
		FileCopy($phWorkSpace & "\" & $phAlbumArtFile, $phAddEncodePath & "\", $FC_OVERWRITE)
	EndIf

   ; clean up
   FileDelete($phWorkSpace & "\*.log")
   FileDelete($phWorkSpace & "\" & $phAlbumArtFile)

   MsgBox($MB_ICONINFORMATION, "", lang("EAH_DONE_TAG"), 5)

EndFunc

; rename files for multi CD sets
; ph<var> is <var> with placeholders replaced
Func multiCdFileRename()

   getFormFields()

   Local $phWorkSpace = replacePlaceHolders($workSpace, $albumInfo)
   Local $fileName, $fileHandle
   Local $format = "%d"

   ; consistency checks
   If Not checkExists($phWorkSpace) Then Return

   ; check if it is a multi CD
   If $albumTotalCDs == 1 Then
	  MsgBox($MB_OK + $MB_ICONWARNING, lang("EAH_MULTICD_MEMBER_ERR"), lang("EAH_MULTICD_NOMEMBER"))
	  Return
   EndIf

	; format of preceding CD number digit (01-01 first track of first CD.mp3 for MultiCD sets with more than 9 CDs)
   If $albumTotalCDs >= 10 Then $format = "%02d"

   ; rename encoded files
   If $encodeEnabled Then
	  $fileHandle = FileFindFirstFile($phWorkSpace & "\*." & $encodeExt)
	  If $fileHandle <> -1 Then
		 While 1
			$fileName = FileFindNextFile($fileHandle)
			If @error Then ExitLoop

			; rename and add to table (catch files which start with "01 " only)
			If checkTrackFile($fileName) Then
			   If Not FileMove($phWorkSpace & "\" & $fileName, $phWorkSpace & "\" & StringFormat($format, $albumCDnumber) & "-" & $fileName) Then
				  If Not MsgBox($MB_OKCANCEL + $MB_ICONWARNING, $phWorkSpace & "\" & $fileName, lang("EAH_MULTICD_RENAME_ERR")) == $IDOK Then Return
			   EndIf
			   Call($debugger, "File " & $fileName & " renamed")
			EndIf
		 WEnd
	  EndIf
	  FileClose($fileHandle)
   EndIf

   ; rename additional encoded files
   If $addEncodeEnabled Then
	  $fileHandle = FileFindFirstFile($phWorkSpace & "\*." & $addEncodeExt)
	  If $fileHandle <> -1 Then
		 While 1
			$fileName = FileFindNextFile($fileHandle)
			If @error Then ExitLoop

			; rename and add to table (catch files which start with "01 " only)
			If checkTrackFile($fileName) Then
			   If Not FileMove($phWorkSpace & "\" & $fileName, $phWorkSpace & "\" & StringFormat($format, $albumCDnumber) & "-" & $fileName) Then
				  If Not MsgBox($MB_OKCANCEL + $MB_ICONWARNING, $phWorkSpace & "\" & $fileName, lang("EAH_MULTICD_RENAME_ERR")) == $IDOK Then Return
			   EndIf
			   Call($debugger, "File " & $fileName & " renamed")
			EndIf
		 WEnd
	  EndIf
	  FileClose($fileHandle)
   EndIf

   ; find log files and rename them
   If $encodeEnabled Or $addEncodeEnabled Then
	  $fileHandle = FileFindFirstFile($phWorkSpace & "\*.log")
	  If $fileHandle <> -1 Then
		 While 1
			$fileName = FileFindNextFile($fileHandle)
			If @error Then ExitLoop

			; rename and add to table (catch files which start with a non-digit character only)
			If checkInfoFile($fileName) Then
			   If Not FileMove($phWorkSpace & "\" & $fileName, $phWorkSpace & "\" & StringFormat($format, $albumCDnumber) & "-" & $fileName) Then
				  If Not MsgBox($MB_OKCANCEL + $MB_ICONWARNING, $phWorkSpace & "\" & $fileName, lang("EAH_MULTICD_RENAME_ERR")) == $IDOK Then Return
			   EndIf
			   Call($debugger, "File " & $fileName & " renamed")
			EndIf
		 WEnd
	  EndIf
	  FileClose($fileHandle)
   EndIf

   MsgBox($MB_ICONINFORMATION, "", lang("EAH_DONE_MULTI"), 5)

EndFunc

; set 'Various Artists' as $albumartist
; use 'ControlSend' because 'ControlSetText' does not trigger EAC to update internal artist variable
Func setArtist($newArtist)
	ControlSetText($windowTitle, "", $formFieldArtist, "") ; remove text in control
	ControlSend($windowTitle, "", $formFieldArtist, $newArtist & "{ENTER}", $SEND_DEFAULT)
	Call($debugger, lang("EAH_MENU_SET_VA"))
EndFunc

; main
; ----------------------------------------------------------------------------------------------------
initialize()

; Exit if EAH and/or EAC is already running
If WinExists($windowTitle) Then
   MsgBox($MB_OK + $MB_ICONWARNING, lang("EAH_RUNNING_ERR"), lang("EAH_RUNNING_DESC"))
   Exit
EndIf

guiMain()
terminate()

Exit

; main program GUI
; ----------------------------------------------------------------------------------------------------
Func guiMain()

	Local $discState
	Local $discPrevState = ""
	Local $vaCheckState
	Local $menuItem = ""
	Local $metaArtist = $vaName

	Run($eacExe)

	; wait for window to become active
	WinWaitActive($windowTitle)
	$windowHandle = WinGetHandle($windowTitle)
	insertMenu()
	createFormFields()

	While 1

		; change window title if necessary
		If WinActive($windowTitle) And Not WinActive($windowEahTitle) Then WinSetTitle($windowHandle, "", $windowEahTitle & " / " & WinGetTitle($windowHandle))

		; check disc change and reset va
		$discState = ControlGetText($windowTitle, "", $formFieldAlbum)
		If $discState <> $discPrevState And $discState <> "" Then
			If ControlGetText($windowTitle, "", $formFieldArtist) == $vaName Then
				$vaState = 1
			Else
				$vaState = 0
			EndIf
			$discPrevState = $discState
		EndIf

		; set state of va check box to make sure it's consistent
		If $vaState Then
			_GUICtrlButton_SetCheck($vaHandle, $BST_CHECKED)
		Else
			_GUICtrlButton_SetCheck($vaHandle, $BST_UNCHECKED)
		EndIf

		; Check state of menubar
		$aInfo = _GUICtrlMenu_GetMenuBarInfo($windowHandle)

		; If menu is active
		If $aInfo[6] = True Then

			; Run through our added items to see if they are active
			For $i = 0 To UBound($command, $UBOUND_ROWS) - 1
				If _GUICtrlMenu_GetItemHighlighted($menuHandle, $command[$i][1], False) Then

					; if active, get text
					If $command[$i][0] <> $menuItem Then $menuItem = $command[$i][2]
				EndIf
			Next
		EndIf

		; menu item selected
		If $menuItem <> "" Then

			; mouse clicked or enter pressed
			If _IsPressed("01", $dll) Or _IsPressed("0D", $dll) Then

				Call($menuItem)

				; reset
				$menuItem = ""

				; wait until mouse/enter released
				While _IsPressed("0D", $dll) Or _IsPressed("01", $dll)
					Sleep(10)
				WEnd
			EndIf

		; check F1 key pressed
		ElseIf _IsPressed(70) Then
			WM_HELP()

		; va check box
		Else
			$vaCheckState = _GUICtrlButton_GetState($vaHandle)

			; run only if check box is pushed
			If BitAND($vaCheckState, $BST_PUSHED) Then
				If BitAND($vaCheckState, $BST_CHECKED) Then
					$vaState = 0
					setArtist($metaArtist)
				Else
					$vaState = 1
					$metaArtist = ControlGetText($windowTitle, "", $formFieldArtist)
					setArtist($vaName)
				EndIf
			EndIf
		EndIf

		Sleep(50)

		; Exit if EAC is closed
		If Not WinExists($windowTitle) Then Return

	WEnd
EndFunc
