#cs
VAH - Verify Audio Helper
AutoIt v3.3.14.5 - http://www.autoitscript.com

Copyright (c) 2020, Bernard Condrau.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
----------------------------------------------------------------------------------------------------
Use this script to verify 2 different directory trees created with EAH and AAH are in sync
- Exact Audio Helper (EAH) for Exact Audio Copy
- Audacity Audio Helper (AAH)
The script detects missing files due to errors in creation and / or moving of files into directories
----------------------------------------------------------------------------------------------------
Note on debugging this script
- Set Globals $debug = 1 and $breakPointExe
- #include <Debug.au3>
- $breakPointExe points to a local DOS command batch file on your computer with following content
  @ECHO OFF
  ECHO Break Point: %1
  PAUSE
- Whenever you call function breakPoint(), the script pauses until hit a key in the DOS batch File
  This helps trouble shoot file modification, copy, and move commands
----------------------------------------------------------------------------------------------------
#ce

#pragma compile(UPX, True)
#pragma compile(AutoItExecuteAllowed, False)
#pragma compile(FileVersion, 1.0.20.1130)
#pragma compile(ProductName, Verify Audio Helper)
#pragma compile(ProductVersion, 1.0)
#pragma compile(LegalCopyright, (c) 2020 Bernard Condrau)
#pragma compile(Comments, GNU General Public License version 3)
#pragma compile(OriginalFilename, VAH.exe)

#include <GUIConstantsEx.au3>
#include <MsgBoxConstants.au3>
#include <Array.au3>
#include <File.au3>


; set debug=0 for final build
Global $debug = 0
Global $breakPointExe = '"C:\PortableCustom\EAH\tools\breakpoint.bat"'
;#include <Debug.au3>

; app
; ----------------------------------------------------------------------------------------------------
Global $appName    = "Verify Audio Helper"
Global $appVersion = "1.0.20.1130" ; same as #pragma above
Global $appDate    = "30 Nov 2020"
Global $appHelpURL = "https://wiki.condrau.com/media:eah"
Global $appHelpURL2= "https://wiki.condrau.com/media:aah"
Global $appAbout   = " VAH..."

; globals
; ----------------------------------------------------------------------------------------------------
Global $formFieldEncodePath, $formFieldAddEncodePath, $formFieldEncodeExt, $formFieldAddEncodeExt
Global $formFieldEncodePathHandle, $formFieldAddEncodePathHandle, $formFieldEncodeExtHandle, $formFieldAddEncodeExtHandle
Global $encodeButtonHandle, $addEncodeButtonHandle


Global $scriptName = StringTrimRight(@ScriptName, 4)
Global $iniFile = @ScriptDir & "\" & $scriptName & ".ini"
Global $langFile = @ScriptDir & "\" & $scriptName & ".en.ini"
Global $menuCnt, $menuHandle, $menuItem  = ""
Global $windowHandle, $windowTitle = $appName
Global $language,$registerHotKeys=True

; debugging
; ----------------------------------------------------------------------------------------------------
If $debug Then
	; avoid compile error missing function _DebugSetup()
	$debugger = "_DebugSetup"
	Call($debugger)
	$debugger = "_DebugReport"
	Call($debugger, "AutoIt runs x64 = " & @AutoItX64)
Else
	$debugger = "noDebugReport"
EndIf

Func noDebugReport($sData, $bLastError = False, $bExit = False)
   ; this function replaces _DebugReport when debugging is off, and does nothing
EndFunc

Func breakPoint($position = @scriptName)
   If $debug Then RunWait($breakPointExe & " " & $position)
EndFunc

; helpers
; ----------------------------------------------------------------------------------------------------
Func registerHotKey($cmd, $text = 0)
   If $registerHotKeys Then
	  Local $hotKey[4][3] = [ _
		 ["check", "{F4}", @TAB & "F4"], _
		 ["exit", "", @TAB & "Alt-F4"] _ ; Alt-F4 is the predefined exit key
	  ]
	  For $i = 0 To UBound($hotKey, $UBOUND_ROWS) - 1
		 If $hotKey[$i][0] == $cmd Then Return $hotKey[$i][1+$text]
	  Next
   EndIf
   Return ""
EndFunc

Func insertMenu()

   Local $itemHandle, $itemText, $newItemText, $accessKey, $accessKeyPos

   ; set menu text and command IDs
   Global $command[4][3] = [ _
	  [lang("VAH_MENU_CHECK"), 0x2000, "check"], _
	  ["", -1], _
	  [lang("VAH_MENU_ABOUT") & $appAbout, 0x2400, "about"], _
	  [lang("VAH_MENU_EXIT"), 0x2600, "exit"] _
   ]

   Call($debugger, "Inserting menu")

   $menuHandle = GUICtrlCreateMenu(lang("VAH_MENU_FILES"))

   ; create menu and hotkeys
   For $i = 0 To UBound($command, $UBOUND_ROWS) - 1
	  $command[$i][0] = $command[$i][0] & registerHotKey($command[$i][2], 1)
	  $command[$i][1] = GUICtrlCreateMenuItem($command[$i][0], $menuHandle)
	  HotKeySet(registerHotKey($command[$i][2]), $command[$i][2])
   Next

   Call($debugger, "Inserting menu done")
EndFunc

; create the form fields in window
Func createFormFields()

   Call($debugger, "Create form fields")

   ; info text
   GUICtrlCreateLabel(lang("VAH_INFOTEXT"), 10, 13)

   ; Path 1
   $encodeButtonHandle = GUICtrlCreateButton(lang("VAH_PATH"), 10, 50, 75)
   $formFieldEncodePathHandle = GUICtrlCreateInput($formFieldEncodePath, 100, 51, 450, 22)

   ; Path 2
   $addEncodeButtonHandle = GUICtrlCreateButton(lang("VAH_PATH"), 10, 75, 75)
   $formFieldAddEncodePathHandle = GUICtrlCreateInput($formFieldAddEncodePath, 100, 77, 450, 22)

   ; Extension 1
   Local $formFieldEncodeExtLabel = GUICtrlCreateLabel(lang("VAH_EXT"), 575, 53, 100)
   $formFieldEncodeExtHandle = GUICtrlCreateInput($formFieldEncodeExt, 645, 51, 100, 22)

   ; Extension 2
   Local $formFieldAddEncodeExtLabel = GUICtrlCreateLabel(lang("VAH_EXT"), 575, 78, 100)
   $formFieldAddEncodeExtHandle = GUICtrlCreateInput($formFieldAddEncodeExt, 645, 77, 100, 22)

EndFunc

Func getFormFields()

   Call($debugger, "Get form field values")

   ; get form fields
   $formFieldEncodePath = GUICtrlRead($formFieldEncodePathHandle) ; mandatory
   $formFieldAddEncodePath = GUICtrlRead($formFieldAddEncodePathHandle) ; mandatory
   $formFieldEncodeExt = GUICtrlRead($formFieldEncodeExtHandle) ; cannot be empty
   $formFieldAddEncodeExt = GUICtrlRead($formFieldAddEncodeExtHandle) ; cannot be empty

EndFunc

Func iniFileRead($filename, $section, $key, $default)

   Local $value

   $value = IniRead($filename, $section, $key, $default)
   If $value == "" Then $value = $default
   Return $value

EndFunc

Func initialize()

   Global $dll = DllOpen("user32.dll")
   Local $myLanguage

   Call($debugger, "initialize")

   ; check whether my language translation file exists
   $myLanguage = IniFileRead($iniFile, "Settings", "Language", "en")
   If FileExists(@ScriptDir & "\" & $scriptName & "." & $myLanguage & ".ini") Then
	  $language = $myLanguage
	  $langFile = @ScriptDir & "\" & $scriptName & "." & $language & ".ini"
   EndIf

   ; read ini file - Settings
   $formFieldEncodePath = IniFileRead($iniFile, "Settings", "EncodePath", "")
   $formFieldAddEncodePath = IniFileRead($iniFile, "Settings", "AddEncodePath", "")
   $formFieldEncodeExt = IniFileRead($iniFile, "Settings", "EncodeExt", "")
   $formFieldAddEncodeExt = IniFileRead($iniFile, "Settings", "AddEncodeExt", "")

   ; remove trailing \
   trimSlash($formFieldEncodePath)
   trimSlash($formFieldAddEncodePath)

   Call($debugger, 'Reading ' & $iniFile & '...' & @CRLF & $language & ' (' & lang("VAH_LANGUAGE") & ') - ' & $langFile & @CRLF & $formFieldEncodePath & @CRLF & $formFieldEncodeExt & @CRLF & $formFieldAddEncodePath & @CRLF & $formFieldAddEncodeExt)
   Call($debugger, "Register hotkeys=" & $registerHotKeys)


EndFunc

Func lang($key)
   $str = IniRead($langFile, $language, $key, $key)
   Return StringReplace(StringReplace($str, "%TAB%", @TAB), "%CRLF%", @CRLF)
EndFunc

Func terminate()
   Call($debugger, 'terminate')
   DllClose($dll)
EndFunc

; file name manipulation
; ----------------------------------------------------------------------------------------------------
Func checkExists($path)
	Local $langMsg
	$path = StringRegExpReplace($path, '^"|"$', "")
	If $path <> "" And Not FileExists($path) Then
		$langMsg = "VAH_PATH_NOT_FOUND"
		If StringRegExp($path, '.*\.exe$') Then $langMsg = "VAH_FILE_NOT_FOUND"
		MsgBox($MB_OK + $MB_ICONWARNING, $path, lang($langMsg))
		Return False
	EndIf
	Return True
EndFunc

Func baseName($name)
	Local $parts = StringSplit($name, ".")
	Local $ext = _ArrayPop($parts)
	Local $skip = $ext <> $formFieldEncodeExt And $ext <> $formFieldAddEncodeExt
	If $skip Then
		Return $name
	Else
		Return _ArrayToString($parts, ".", 1)
	EndIf
EndFunc

Func trimSlash(ByRef $path)
   If StringRight($path, 1) == '\' Then $path = StringTrimRight($path, 1)
EndFunc

; progress window
; ----------------------------------------------------------------------------------------------------
Func progress($percent, $artist = 0, $artistCount = 0, $album = 0, $albumCount = 0)

	Static Local $previous = -1

	If $percent == 0 Then

		; calculate position of progress window
		Local $winPos = WinGetPos($windowHandle)
		Local $desktop = WinGetPos("Program Manager")
		Local $xPos = $winPos[0] + $winPos[2]/2 - 150
		Local $yPos = $winPos[1] - 120
		If $yPos < 0 Then $yPos = 0
		If $winPos[1] < 120 And $winPos[1] + $winPos[3] + 120 < $desktop[3] - 160 Then $yPos = $winPos[1] + $winPos[3] + 20

		; initialize progress window
		ProgressOn("", lang("VAH_PROGRESS"), lang("VAH_PROGRESS_PERCENT") & ": 0%" & @CRLF & lang("VAH_PROGRESS_ARTISTS") & ": 0/0" & @CRLF & lang("VAH_PROGRESS_ALBUMS") & ": 0/0", $xPos, $yPos, BitOR($DLG_NOTITLE, $DLG_NOTONTOP))
	Else

		; update progress window
		$percent = Round($percent, 1)
		If $percent <> $previous Then
			ProgressSet($percent, lang("VAH_PROGRESS_PERCENT") & ": " & $percent & "%" & @CRLF & lang("VAH_PROGRESS_ARTISTS") & ": " & $artist & "/" & $artistCount & @CRLF & lang("VAH_PROGRESS_ALBUMS") & ": " & $album & "/" & $albumCount)
			$previous = $percent
		EndIf

		; reset $previous
		If $previous == 100 Then $previous = -1
	EndIf
EndFunc

; actions
; ----------------------------------------------------------------------------------------------------
; about msg box
Func about()
	MsgBox($MB_ICONINFORMATION, lang("VAH_MENU_ABOUT") & " " & $appName, $appName & " v" & $appVersion & " (" & $appDate & ")" & @CRLF & $appHelpURL & @CRLF & $appHelpURL2, 10)
EndFunc

; check consistency of folders and files
Func check()

	getFormFields()

	If Not $formFieldEncodePath Or Not $formFieldAddEncodePath Or Not $formFieldEncodeExt Or Not $formFieldAddEncodeExt Then
		MsgBox($MB_OK + $MB_ICONERROR, lang("VAH_MANDATORY"), lang("VAH_MANDATORY_DESC"))
		Return
	EndIf

	; consistency checks
	If Not checkExists($formFieldEncodePath) Then Return
	If Not checkExists($formFieldAddEncodePath) Then Return

	Local $errorList[10] = [ _
		"", _
		lang("VAH_PATH_NOT_FOUND"), _
		lang("VAH_UNKNOWN_ERROR") & " #2", _
		lang("VAH_UNKNOWN_ERROR") & " #3", _
		lang("VAH_UNKNOWN_ERROR") & " #4", _
		lang("VAH_UNKNOWN_ERROR") & " #5", _
		lang("VAH_UNKNOWN_ERROR") & " #6", _
		lang("VAH_UNKNOWN_ERROR") & " #7", _
		lang("VAH_UNKNOWN_ERROR") & " #8", _
		lang("VAH_NO_FILES") _
	]
	Local $encodeBaseName = ""
	Local $encodePrevName = ""
	Local $addEncodeBaseName = ""
	Local $addEncodePrevName = ""
	Local $evaluateTracks = False
	Local $evaluateAlbums = False
	Local $evaluateArtists = False

	progress(0)

	$evaluateArtists = True

	Local $encodeArtists = _FileListToArrayRec($formFieldEncodePath, "*", $FLTAR_FOLDERS, $FLTAR_NORECUR, $FLTAR_SORT, $FLTAR_NOPATH)
	If @error Then
		MsgBox($MB_OK + $MB_ICONERROR, $errorList[@extended], $formFieldEncodePath)
		$evaluateArtists = False
	EndIf
	If $debug And $evaluateArtists Then _ArrayDisplay ($encodeArtists, lang("VAH_ARTIST_FOLDERS") & " " & $formFieldEncodePath)

	Local $addEncodeArtists = _FileListToArrayRec($formFieldAddEncodePath, "*", $FLTAR_FOLDERS, $FLTAR_NORECUR, $FLTAR_SORT, $FLTAR_NOPATH)
	If @error Then
		MsgBox($MB_OK + $MB_ICONERROR, $errorList[@extended], $formFieldAddEncodePath)
		$evaluateArtists = False
	EndIf
	If $debug And $evaluateArtists Then _ArrayDisplay ($addEncodeArtists, lang("VAH_ARTIST_FOLDERS") & " " & $formFieldAddEncodePath)

	If $evaluateArtists Then

		; top dir (artists)
		Local $equal = True
		Local $artistCount = $encodeArtists[0]
		If $addEncodeArtists[0] < $artistCount Then
			$artistCount = $addEncodeArtists[0]
			$equal = False
		EndIf
		Local $position = $artistCount
		For $i = 1 To $artistCount
			If $encodeArtists[$i] <> $addEncodeArtists[$i] Then
				$position = $i
				$equal = False
				ExitLoop
			EndIf
		Next
		If Not $equal Then
			Call($debugger, 'pos: ' & $position & ', size=' & $encodeArtists[0] & '<->' & $addEncodeArtists[0])
			Call($debugger, 'enc: ' & $encodeArtists[$position])
			Call($debugger, 'add: ' & $addEncodeArtists[$position])
			$id = MsgBox($MB_ABORTRETRYIGNORE + $MB_ICONWARNING + $MB_DEFBUTTON1, lang("VAH_ARTIST_TITLE"), lang("VAH_ARTIST_DESC") & ": " & $position & @CRLF & lang("VAH_ARTIST") & ": " & $encodeArtists[$position] & @CRLF & lang("VAH_ARTIST") & ": " & $addEncodeArtists[$position] & @CRLF & @CRLF & lang("VAH_MSGBOX_ARTIST"))
			If $id == $IDABORT Then Return
			If $id == $IDRETRY Then
				_ArrayDisplay ($encodeArtists, lang("VAH_ARTIST_FOLDERS") & " " & $formFieldEncodePath)
				_ArrayDisplay ($addEncodeArtists, lang("VAH_ARTIST_FOLDERS") & " " & $formFieldAddEncodePath)
			EndIf
		EndIf

		Local $encodeArtistsPath = _FileListToArrayRec($formFieldEncodePath, "*", $FLTAR_FOLDERS, $FLTAR_NORECUR, $FLTAR_SORT, $FLTAR_FULLPATH)
		Local $addEncodeArtistsPath = _FileListToArrayRec($formFieldAddEncodePath, "*", $FLTAR_FOLDERS, $FLTAR_NORECUR, $FLTAR_SORT, $FLTAR_FULLPATH)

		; 2nd level dir (albums)
		For $i = 1 To $artistCount
			If $encodeArtists[$i] <> $addEncodeArtists[$i] Then
				$id = MsgBox($MB_ABORTRETRYIGNORE + $MB_ICONWARNING + $MB_DEFBUTTON1, lang("VAH_ARTIST_TITLE"), lang("VAH_ARTIST_DESC") & ": " & $i & @CRLF & lang("VAH_ARTIST") & ": " & $encodeArtists[$i] & @CRLF & lang("VAH_ARTIST") & ": " & $addEncodeArtists[$i] & @CRLF & @CRLF & lang("VAH_MSGBOX_ARTIST"))
				If $id == $IDABORT Then Return
				If $id == $IDRETRY Then
					_ArrayDisplay ($encodeArtists, lang("VAH_ARTIST_FOLDERS") & " " & $formFieldEncodePath)
					_ArrayDisplay ($addEncodeArtists, lang("VAH_ARTIST_FOLDERS") & " " & $formFieldAddEncodePath)
				EndIf
			ElseIf $encodeArtists[$i] <> "." And $encodeArtists[$i] <> ".." Then
				$evaluateAlbums = True

				Local $encodeAlbums = _FileListToArrayRec($encodeArtistsPath[$i], "*", $FLTAR_FOLDERS, $FLTAR_NORECUR, $FLTAR_SORT, $FLTAR_NOPATH)
				If @error Then
					MsgBox($MB_OK + $MB_ICONWARNING, $errorList[@extended], $encodeArtistsPath[$i])
					$evaluateAlbums = False
				EndIf
				If $debug And $evaluateAlbums Then _ArrayDisplay ($encodeAlbums, lang("VAH_ALBUM_FOLDERS") & " " & $encodeArtistsPath[$i])

				Local $addEncodeAlbums = _FileListToArrayRec($addEncodeArtistsPath[$i], "*", $FLTAR_FOLDERS, $FLTAR_NORECUR, $FLTAR_SORT, $FLTAR_NOPATH)
				If @error Then
					MsgBox($MB_OK + $MB_ICONWARNING, $errorList[@extended], $addEncodeArtistsPath[$i])
					$evaluateAlbums = False
				EndIf
				If $debug And $evaluateAlbums Then _ArrayDisplay ($addEncodeAlbums, lang("VAH_ALBUM_FOLDERS") & " " & $addEncodeArtistsPath[$i])

				If $evaluateAlbums Then
					$equal = True
					Local $albumCount = $encodeAlbums[0]
					If $addEncodeAlbums[0] < $albumCount Then
						$albumCount = $addEncodeAlbums[0]
						$equal = False
					EndIf
					$position = $albumCount
					For $j = 1 To $albumCount
						If $encodeAlbums[$j] <> $addEncodeAlbums[$j] Then
							$position = $j
							$equal = False
							ExitLoop
						EndIf
					Next

					If Not $equal Then
						Call($debugger, 'pos: ' & $position & ', size=' & $encodeAlbums[0] & '<->' & $addEncodeAlbums[0])
						Call($debugger, 'enc: ' & $encodeAlbums[$position])
						Call($debugger, 'add: ' & $addEncodeAlbums[$position])
						$id = MsgBox($MB_ABORTRETRYIGNORE + $MB_ICONWARNING + $MB_DEFBUTTON1, lang("VAH_ALBUM_TITLE"), lang("VAH_ALBUM_DESC") & ": " & $position & @CRLF & lang("VAH_ALBUM") & ": " & $encodeArtists[$i] & " * " & $encodeAlbums[$position] & @CRLF & lang("VAH_ALBUM") & ": " & $addEncodeArtists[$i] & " * " & $addEncodeAlbums[$position] & @CRLF & @CRLF & lang("VAH_MSGBOX_ALBUM"))
						If $id == $IDABORT Then Return
						If $id == $IDRETRY Then
							_ArrayDisplay ($encodeAlbums, lang("VAH_ALBUM_FOLDERS") & " " & $encodeArtistsPath[$i])
							_ArrayDisplay ($addEncodeAlbums, lang("VAH_ALBUM_FOLDERS") & " " & $addEncodeArtistsPath[$i])
							$evaluateAlbums = False
						EndIf
					EndIf
				EndIf

				If $evaluateAlbums Then
					$encodeAlbumsPath = _FileListToArrayRec($encodeArtistsPath[$i], "*", $FLTAR_FOLDERS, $FLTAR_NORECUR, $FLTAR_SORT, $FLTAR_FULLPATH)
					$addEncodeAlbumsPath = _FileListToArrayRec($addEncodeArtistsPath[$i], "*", $FLTAR_FOLDERS, $FLTAR_NORECUR, $FLTAR_SORT, $FLTAR_FULLPATH)

					; 3rd level files (tracks)
					For $j = 1 To $albumCount
						If $encodeAlbums[$j] <> $addEncodeAlbums[$j] Then
							$id = MsgBox($MB_ABORTRETRYIGNORE + $MB_ICONWARNING + $MB_DEFBUTTON1, lang("VAH_ALBUM_TITLE"), lang("VAH_ALBUM_DESC") & ": " & $j & @CRLF & lang("VAH_ALBUM") & ": " & $encodeArtists[$i] & " * " & $encodeAlbums[$j] & @CRLF & lang("VAH_ALBUM") & ": " & $addEncodeArtists[$i] & " * " & $addEncodeAlbums[$j] & @CRLF & @CRLF & lang("VAH_MSGBOX_ALBUM"))
							If $id == $IDABORT Then Return
							If $id == $IDRETRY Then
								_ArrayDisplay ($encodeAlbums, lang("VAH_ALBUM_FOLDERS") & " " & $encodeArtistsPath[$i])
								_ArrayDisplay ($addEncodeAlbums, lang("VAH_ALBUM_FOLDERS") & " " & $addEncodeArtistsPath[$i])
							EndIf
						ElseIf $encodeAlbums[$j] <> "." And $encodeAlbums[$j] <> ".." Then
							$evaluateTracks = True

							Local $encodeTracks = _FileListToArrayRec($encodeAlbumsPath[$j], "*", $FLTAR_FILES, $FLTAR_NORECUR, $FLTAR_SORT, $FLTAR_NOPATH)
							If @error Then
								MsgBox($MB_OK + $MB_ICONWARNING, $errorList[@extended], $encodeAlbumsPath[$j])
								$evaluateTracks = False
							EndIf
							If $debug And $evaluateTracks Then _ArrayDisplay ($encodeTracks, lang("VAH_TRACK_FILES") & " " & $encodeAlbumsPath[$j])

							Local $addEncodeTracks = _FileListToArrayRec($addEncodeAlbumsPath[$j], "*", $FLTAR_FILES, $FLTAR_NORECUR, $FLTAR_SORT, $FLTAR_NOPATH)
							If @error Then
								MsgBox($MB_OK + $MB_ICONWARNING, $errorList[@extended], $addEncodeAlbumsPath[$j])
								$evaluateTracks = False
							EndIf
							If $debug And $evaluateTracks Then _ArrayDisplay ($addEncodeTracks, lang("VAH_TRACK_FILES") & " " & $addEncodeAlbumsPath[$j])

							If $evaluateTracks Then
								If $encodeTracks[0] <> $addEncodeTracks[0] Then
									Call($debugger, $encodeArtists[$i] & " * " & $encodeAlbums[$j] & " * " & 'track number: ' & $encodeTracks[0] & '<->' & $addEncodeTracks[0])
									$id = MsgBox($MB_ABORTRETRYIGNORE + $MB_ICONWARNING + $MB_DEFBUTTON1, lang("VAH_FILE_COUNT_TITLE"), $encodeArtists[$i] & " * " & $encodeAlbums[$j] & " * " & lang("VAH_FILE_COUNT_DESC") & ": " & $encodeTracks[0] & @CRLF & $addEncodeArtists[$i] & " * " & $addEncodeAlbums[$j] & " * " & lang("VAH_FILE_COUNT_DESC") & ": " & $addEncodeTracks[0] & @CRLF & @CRLF & lang("VAH_MSGBOX_TRACK"))
									If $id == $IDABORT Then Return
									If $id == $IDRETRY Then
										_ArrayDisplay ($encodeTracks, lang("VAH_TRACK_FILES") & " " & $encodeAlbumsPath[$j])
										_ArrayDisplay ($addEncodeTracks, lang("VAH_TRACK_FILES") & " " & $addEncodeAlbumsPath[$j])
										$evaluateTracks = False
									EndIf
								EndIf
							EndIf

							If $evaluateTracks Then

								Local $fileCount = $encodeTracks[0]
								If $addEncodeTracks[0] < $fileCount Then
									$fileCount = $addEncodeTracks[0]
								EndIf
								For $k = 1 To $fileCount
									$encodeBaseName = baseName($encodeTracks[$k])
									$addEncodeBaseName = baseName($addEncodeTracks[$k])
									If $encodeBaseName <> $addEncodeBaseName Then
										Call($debugger, 'enc: ' & $encodeArtists[$i] & " * " & $encodeAlbums[$j] & " * " & $encodeTracks[$k])
										Call($debugger, 'add: ' & $addEncodeArtists[$i] & " * " & $addEncodeAlbums[$j] & " * " & $addEncodeTracks[$k])
										$id = MsgBox($MB_ABORTRETRYIGNORE + $MB_ICONWARNING + $MB_DEFBUTTON1, lang("VAH_TRACK_TITLE"), lang("VAH_TRACK_DESC") & ": " & $k & @CRLF & $encodeArtists[$i] & " * " & $encodeAlbums[$j] & " * " & $encodeTracks[$k] & @CRLF & $addEncodeArtists[$i] & " * " & $addEncodeAlbums[$j] & " * " & $addEncodeTracks[$k] & @CRLF & @CRLF & lang("VAH_MSGBOX_TRACK"))
										If $id == $IDABORT Then Return
										If $id == $IDRETRY Then
											_ArrayDisplay ($encodeTracks, lang("VAH_TRACK_FILES") & " " & $encodeAlbumsPath[$j])
											_ArrayDisplay ($addEncodeTracks, lang("VAH_TRACK_FILES") & " " & $addEncodeAlbumsPath[$j])
											ExitLoop
										EndIf
									EndIf
									If $encodeBaseName == $encodePrevName Then
										$id = MsgBox($MB_ABORTRETRYIGNORE + $MB_ICONWARNING + $MB_DEFBUTTON1, lang("VAH_TRACK_TITLE"), lang("VAH_TRACK_SAME_DESC") & ": " & $k & @CRLF & $encodeArtists[$i] & " * " & $encodeAlbums[$j] & " * " & $encodeTracks[$k] & @CRLF & $encodeArtists[$i] & " * " & $encodeAlbums[$j] & " * " & $encodeTracks[$k-1] & @CRLF & @CRLF & lang("VAH_MSGBOX_TRACK"))
										If $id == $IDABORT Then Return
										If $id == $IDRETRY Then
											_ArrayDisplay ($encodeTracks, lang("VAH_TRACK_FILES") & " " & $encodeAlbumsPath[$j])
											ExitLoop
										EndIf
									EndIf
									If $addEncodeBaseName == $addEncodePrevName Then
										$id = MsgBox($MB_ABORTRETRYIGNORE + $MB_ICONWARNING + $MB_DEFBUTTON1, lang("VAH_TRACK_TITLE"), lang("VAH_TRACK_SAME_DESC") & ": " & $k & @CRLF & $addEncodeArtists[$i] & " * " & $addEncodeAlbums[$j] & " * " & $addEncodeTracks[$k] & @CRLF & $addEncodeArtists[$i] & " * " & $addEncodeAlbums[$j] & " * " & $addEncodeTracks[$k-1] & @CRLF & @CRLF & lang("VAH_MSGBOX_TRACK"))
										If $id == $IDABORT Then Return
										If $id == $IDRETRY Then
											_ArrayDisplay ($addEncodeTracks, lang("VAH_TRACK_FILES") & " " & $addEncodeAlbumsPath[$j])
											ExitLoop
										EndIf
									EndIf
									$encodePrevName = $encodeBaseName
									$addEncodePrevName = $addEncodeBaseName
								Next
							EndIf ; $evaluateTracks
						EndIf
						progress($i/$artistCount*100, $i, $artistCount, $j, $albumCount)
					Next
				EndIf ; $evaluateAlbums
			EndIf
		Next
	EndIf ; $evaluateArtists
	If $evaluateArtists And $evaluateAlbums And $evaluateTracks Then progress(100, $artistCount, $artistCount, $albumCount, $albumCount)
EndFunc

; main
; ----------------------------------------------------------------------------------------------------
initialize()

; Exit if VAH is already running
If WinExists($windowTitle) Then
   MsgBox($MB_OK + $MB_ICONWARNING, lang("VAH_RUNNING_ERR"), lang("VAH_RUNNING_DESC"))
   Exit
EndIf

guiMain()
terminate()

Exit

; main program GUI
; ----------------------------------------------------------------------------------------------------
Func guiMain()

	Local $path

	$windowHandle = GUICreate($windowTitle, 800, 150)
	insertMenu()
	createFormFields()

	GUISetState()

	While 1
		$cmd = GUIGetMsg()
		If $cmd == $GUI_EVENT_CLOSE Or $cmd == $command[UBound($command, $UBOUND_ROWS) - 1][1] Then
			Exit
		ElseIf $cmd == $encodeButtonHandle Then
			$path = FileSelectFolder(lang("VAH_SELECT_FOLDER"), $formFieldEncodePath)
            If $path Then GUICtrlSetData($formFieldEncodePathHandle, $path)
		ElseIf $cmd == $addEncodeButtonHandle Then
            $path = FileSelectFolder(lang("VAH_SELECT_FOLDER"), $formFieldAddEncodePath)
            If $path Then GUICtrlSetData($formFieldAddEncodePathHandle, $path)
		Else
			For $i = 0 To UBound($command, $UBOUND_ROWS) - 2
				If $cmd == $command[$i][1] Then
					Call($debugger, "command=" & $command[$i][2])
					Call($command[$i][2])
				EndIf
			Next
		EndIf
		Sleep(50)
	WEnd
EndFunc
